"""Generate code for common functionality shared across the tests."""

import io
import re

from icontract import ensure

from aas_core_codegen.common import (
    Stripped,
    Identifier,
)
from aas_core_codegen.cpp import common as cpp_common
from aas_core_codegen.cpp.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
    INDENT4 as IIII,
)


def _library_namespace_to_upper_snake(
    library_namespace: Stripped,
) -> Identifier:
    """
    Convert a CamelCase dotted namespace into UPPER_SNAKE_CASE.

    >>> _library_namespace_to_upper_snake(Stripped("aas_core::aas_3_0"))
    'AAS_CORE_AAS_3_0'

    >>> _library_namespace_to_upper_snake(Stripped("SimpleTest"))
    'SIMPLE_TEST'

    >>> _library_namespace_to_upper_snake(Stripped("Already_Snake"))
    'ALREADY_SNAKE'
    """
    result = library_namespace.replace("::", "_")

    # Insert underscore between lowercase and uppercase (aasCore → aas_Core)
    result = re.sub(r"(?<=[a-z])(?=[A-Z])", "_", result)

    return Identifier(result.upper())


# fmt: off
@ensure(
    lambda result:
    result.endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate_header(library_namespace: Stripped) -> str:
    """Generate header for common functionality shared across the tests."""
    include_guard_var = cpp_common.include_guard_var(Stripped("test::common"))

    include_prefix_path = cpp_common.generate_include_prefix_path(library_namespace)

    environment_variable = _library_namespace_to_upper_snake(library_namespace)

    blocks = [
        cpp_common.WARNING,
        Stripped(
            f"""\
#ifndef {include_guard_var}
#define {include_guard_var}"""
        ),
        Stripped(
            f"""\
#include <{include_prefix_path}/types.hpp>"""
        ),
        Stripped(
            """\
#include <nlohmann/json.hpp>"""
        ),
        Stripped(
            """\
#include <deque>
#include <filesystem>"""
        ),
        Stripped(
            """\
/**
* Provide common functionalities shared across the tests.
*/
namespace test {
namespace common {"""
        ),
        Stripped(
            f"""\
/**
 * Set to true if we run tests in the record mode according to the environment
 * variable `{environment_variable}_TEST_RECORD_MODE`.
 *
 * \\remark It is tedious to record manually all the expected error messages. Therefore
 * we include this variable to steer the automatic recording. We intentionally
 * intertwine the recording code with the test code to keep them close to each other
 * so that they are easier to maintain.
 */
bool DetermineRecordMode();"""
        ),
        Stripped(
            f"""\
/**
 * Point to the directory with the test data according to the environment
 * variable `{environment_variable}_TEST_DATA_DIR`.
 */
std::filesystem::path DetermineTestDataDir();"""
        ),
        Stripped(
            f"""\
/**
 * Represent `that` instance as a human-readable line of an iteration trace.
 */
std::wstring TraceMark(
{I}const {library_namespace}::types::IClass& that
);"""
        ),
        Stripped(
            f"""\
/**
 * Find files beneath the `root` recursively which have the `suffix`.
 *
 * If the `root` does not exist, return an empty deque.
 *
 * If the `root` is not a directory, throw a runtime error.
 *
 * The files are sorted.
 */
std::deque<std::filesystem::path> FindFilesBySuffixRecursively(
{I}const std::filesystem::path& root,
{I}const std::string& suffix
);"""
        ),
        Stripped(
            f"""\
/**
 * List non-recursively the subdirectories contained in the \\p root.
 *
 * @param root directory that you want to list
 * @return subdirectories beneath \\p root, sorted
 */
std::deque<std::filesystem::path> ListSubdirectories(
{I}const std::filesystem::path& root
);"""
        ),
        Stripped(
            f"""\
/**
 * Check that the content coincides with the file or re-record if in record mode.
 */
void AssertContentEqualsExpectedOrRecord(
{I}const std::string& content,
{I}const std::filesystem::path& path
);"""
        ),
        Stripped(
            f"""\
template<typename T>
std::string JoinStrings(const T& container, const std::string& delimiter) {{
{I}static_assert(std::is_same<typename T::value_type, std::string>::value);

{I}size_t size = 0;
{I}for (const std::string& part : container
{II}) {{
{II}size += part.size();
{I}}}
{I}size += delimiter.size() * (container.size() - 1);

{I}std::string result;
{I}result.reserve(size);

{I}auto it = container.begin();
{I}for (size_t i = 0; i < container.size() - 1; ++i) {{
{II}result.append(*it);
{II}result.append(delimiter);
{II}++it;
{I}}}

{I}result.append(*it);

{I}return result;
}}"""
        ),
        Stripped(
            f"""\
template<typename T>
std::wstring JoinWstrings(const T& container, const std::wstring& delimiter) {{
{I}if (container.size() == 0) {{
{II}return L"";
{I}}}

{I}static_assert(std::is_same<typename T::value_type, std::wstring>::value);

{I}size_t size = 0;
{I}for (const std::wstring& part : container
{II}) {{
{II}size += part.size();
{I}}}
{I}size += delimiter.size() * (container.size() - 1);

{I}std::wstring result;
{I}result.reserve(size);

{I}auto it = container.begin();
{I}for (size_t i = 0; i < container.size() - 1; ++i) {{
{II}result.append(*it);
{II}result.append(delimiter);
{II}++it;
{I}}}

{I}result.append(*it);

{I}return result;
}}"""
        ),
        Stripped(
            f"""\
std::string MustReadString(
{I}const std::filesystem::path& path
);"""
        ),
        Stripped(
            """\
}  // namespace common
}  // namespace test"""
        ),
        Stripped(
            f"""\
#endif  // {include_guard_var}"""
        ),
        cpp_common.WARNING,
    ]

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(block)

    writer.write("\n")

    return writer.getvalue()


# fmt: off
@ensure(
    lambda result:
    result.endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate_implementation(library_namespace: Stripped) -> str:
    """Generate implementation for common functionality shared across the tests."""
    include_prefix_path = cpp_common.generate_include_prefix_path(library_namespace)

    environment_variable = _library_namespace_to_upper_snake(library_namespace)

    blocks = [
        cpp_common.WARNING,
        Stripped(
            '''\
#include "./common.hpp"'''
        ),
        Stripped(
            f"""\
#include <{include_prefix_path}/stringification.hpp>
#include <{include_prefix_path}/wstringification.hpp>"""
        ),
        Stripped(
            """\
#include <catch2/catch.hpp>"""
        ),
        Stripped(
            """\
#include <algorithm>
#include <cstdlib>
#include <cstring>
#include <fstream>
#include <queue>
#include <set>
#include <stdexcept>
#include <type_traits>"""
        ),
        Stripped(
            f"""\
namespace aas = {library_namespace};"""
        ),
        Stripped(
            """\
namespace fs = std::filesystem;"""
        ),
        Stripped(
            """\
namespace test {
namespace common {"""
        ),
        Stripped(
            f"""\
static std::optional<std::string> GetEnv(const char *variable_name) {{
{I}// NOTE (empwilli): MSVC complains about getenv being unsafe.
{I}// This should not be an issue as we don't run our tests in parallel, though,
{I}// so we use getenv_s in Windows. However, this function is a Microsoft
{I}// extension.
#ifdef _WIN32
{I}char buffer[256];
{I}size_t len;
{I}errno_t error = getenv_s(
{II}&len, buffer, sizeof(buffer), variable_name
{I});
{I}if (error != 0) {{
{II}throw std::runtime_error(
{III}aas::common::Concat(
{IIII}"The getenv_s returned an error code ",
{IIII}std::to_string(error),
{IIII}" when asking for the variable ",
{IIII}variable_name
{III})
{II});
{I}}}

{I}if (len == 0) {{
{II}return std::nullopt;
{I}}}

{I}if (buffer[len - 1] != 0) {{
{II}throw std::logic_error(
{III}aas::common::Concat(
{IIII}"Expected the last byte of buffer for getenv_s on ",
{IIII}variable_name,
{IIII}" to be zero, but got ",
{IIII}std::to_string(static_cast<std::uint8_t>(buffer[len - 1]))
{III})
{II});
{I}}}

{I}return std::optional {{ std::string(buffer, len - 1) }};
#else
{I}auto value = getenv(variable_name);

{I}if (value == nullptr) {{
{II}return std::nullopt;
{I}}}

{I}return std::optional {{ std::string(value) }};
#endif
}}"""
        ),
        Stripped(
            f"""\
bool DetermineRecordMode() {{
{I}const char* variable_name = "{environment_variable}_TEST_RECORD_MODE";

{I}auto result = GetEnv(variable_name);

{I}if (!result.has_value())
{I}{{
{II}return false;
{I}}}

{I}static const std::set<std::string> hot_value_set{{
{II}"on", "ON", "On",
{II}"true", "TRUE", "True",
{II}"1",
{II}"yes", "Yes", "YES"
{I}}};
{I}return hot_value_set.find(result.value()) != hot_value_set.end();
}}"""
        ),
        Stripped(
            f"""\
std::filesystem::path DetermineTestDataDir() {{
{I}const char* variable_name = "{environment_variable}_TEST_DATA_DIR";

{I}auto result = GetEnv(variable_name);

{I}if (!result.has_value())
{I}{{
{II}throw std::runtime_error(
{III}aas::common::Concat(
{IIII}"The environment variable ",
{IIII}variable_name,
{IIII}" has not been set."
{III})
{II});
{I}}}

{I}return fs::path(result.value());
}}"""
        ),
        Stripped(
            f"""\
/**
 * Write `text` to the given `path` encoded as UTF-8.
 */
void MustWriteWstringAsUtf8(
{I}const std::filesystem::path& path,
{I}const std::wstring& text
) {{
{I}std::ofstream ofs(path, std::ios::binary);

{I}const std::string encoded = aas::common::WstringToUtf8(text);
{I}ofs.write(encoded.data(), encoded.size());

{I}if (ofs.fail()) {{
{II}throw std::runtime_error(
{III}aas::common::Concat(
{IIII}"Failed to write to ",
{IIII}path.string(),
{IIII}"; the fail bit of the file stream is set"
{III})
{II});
{I}}}

{I}if (ofs.bad()) {{
{II}throw std::runtime_error(
{III}aas::common::Concat(
{IIII}"Failed to write to ",
{IIII}path.string(),
{IIII}"; the bad bit of the file stream is set"
{III})
{II});
{I}}}
}}"""
        ),
        Stripped(
            f"""\
/**
 * Read the content of the file `path` and decode it from UTF-8.
 */
std::wstring MustReadWstringAsUtf8(
{I}const std::filesystem::path& path
) {{
{I}std::ifstream ifs(path, std::ios::binary);
{I}ifs.seekg(0, std::ios::end);
{I}std::streamoff size = ifs.tellg();
{I}if (size < 0) {{
{II}throw std::runtime_error(
{III}aas::common::Concat(
{IIII}"Unexpected negative size of ",
{IIII}path.string(),
{IIII}": ",
{IIII}std::to_string(size)
{III})
{II});
{I}}}

{I}std::string buffer(static_cast<size_t>(size), ' ');
{I}ifs.seekg(0);
{I}ifs.read(&buffer[0], size);

{I}if (ifs.fail()) {{
{II}throw std::runtime_error(
{III}aas::common::Concat(
{IIII}"Failed to read from ",
{IIII}path.string(),
{IIII}"; the fail bit of the file stream is set"
{III})
{II});
{I}}}

{I}if (ifs.bad()) {{
{II}throw std::runtime_error(
{III}aas::common::Concat(
{IIII}"Failed to read from ",
{IIII}path.string(),
{IIII}"; the bad bit of the file stream is set"
{III})
{II});
{I}}}

{I}try {{
{II}return aas::common::Utf8ToWstring(buffer);
{I}}} catch (const std::exception& exception) {{
{II}throw std::runtime_error(
{III}aas::common::Concat(
{IIII}"Failed to de-code ",
{IIII}path.string(),
{IIII}" as UTF-8: ",
{IIII}exception.what()
{III})
{II});
{I}}}
}}"""
        ),
        Stripped(
            f"""\
/**
 * Write `text` to the given `path` as binary.
 */
void MustWriteString(
{I}const std::filesystem::path& path,
{I}const std::string& text
) {{
{I}std::ofstream ofs(path, std::ios::binary);

{I}ofs.write(text.data(), text.size());

{I}if (ofs.fail()) {{
{II}throw std::runtime_error(
{III}aas::common::Concat(
{IIII}"Failed to write to ",
{IIII}path.string(),
{IIII}"; the fail bit of the file stream is set"
{III})
{II});
{I}}}

{I}if (ofs.bad()) {{
{II}throw std::runtime_error(
{III}aas::common::Concat(
{IIII}"Failed to write to ",
{IIII}path.string(),
{IIII}"; the bad bit of the file stream is set"
{III})
{II});
{I}}}
}}"""
        ),
        Stripped(
            f"""\
/**
 * Read the content of the file `path` as binary.
 */
std::string MustReadString(
{I}const std::filesystem::path& path
) {{
{I}std::ifstream ifs(path, std::ios::binary);
{I}ifs.seekg(0, std::ios::end);
{I}std::streamoff size = ifs.tellg();
{I}if (size < 0) {{
{II}throw std::runtime_error(
{III}aas::common::Concat(
{IIII}"Unexpected negative size of ",
{IIII}path.string(),
{IIII}": ",
{IIII}std::to_string(size)
{III})
{II});
{I}}}

{I}std::string buffer(static_cast<size_t>(size), ' ');
{I}ifs.seekg(0);
{I}ifs.read(&buffer[0], size);

{I}if (ifs.fail()) {{
{II}throw std::runtime_error(
{III}aas::common::Concat(
{IIII}"Failed to read from ",
{IIII}path.string(),
{IIII}"; the fail bit of the file stream is set"
{III})
{II});
{I}}}

{I}if (ifs.bad()) {{
{II}throw std::runtime_error(
{III}aas::common::Concat(
{IIII}"Failed to read from ",
{IIII}path.string(),
{IIII}"; the bad bit of the file stream is set"
{III})
{II});
{I}}}

{I}return buffer;
}}"""
        ),
        Stripped(
            f"""\
std::wstring TraceMark(
{I}const {library_namespace}::types::IClass& that
) {{
{I}const std::wstring model_type = aas::wstringification::to_wstring(
{II}that.model_type()
{I});
{I}return model_type;
}}"""
        ),
        Stripped(
            f"""\
bool StringEndsWith(
{I}const std::string& text,
{I}const std::string& suffix
) {{
{I}if (text.size() < suffix.size()) {{
{II}return false;
{I}}}

{I}return (
{II}text.substr(text.size() - suffix.size(), suffix.size())
{III}== suffix
{I});
}}"""
        ),
        Stripped(
            f"""\
std::deque<std::filesystem::path> FindFilesBySuffixRecursively(
{I}const std::filesystem::path& root,
{I}const std::string& suffix
) {{
{I}if (!fs::exists(root)) {{
{II}return std::deque<std::filesystem::path>();
{I}}}

{I}if (!fs::is_directory(root)) {{
{II}throw std::runtime_error(
{III}aas::common::Concat(
{IIII}"Expected to search for files with suffix recursively in ",
{IIII}root.string(),
{IIII}", but it is not a directory"
{III})
{II});
{I}}}

{I}std::deque<fs::path> result;

{I}for (
{II}const fs::directory_entry& entry
{II}: fs::recursive_directory_iterator(root)
{II}) {{
{II}if (
{III}StringEndsWith(
{IIII}entry.path().filename().string(),
{IIII}suffix
{III})
{III}) {{
{III}result.push_back(entry.path());
{II}}}
{I}}}

{I}std::sort(result.begin(), result.end());

{I}return result;
}}"""
        ),
        Stripped(
            f"""\
std::deque<std::filesystem::path> ListSubdirectories(
{I}const std::filesystem::path& root
) {{
{I}if (!fs::exists(root)) {{
{II}throw std::runtime_error(
{III}aas::common::Concat(
{IIII}"The root directory which you wanted to list for subdirectories "
{IIII}"does not exist: ",
{IIII}root.string()
{III})
{II});
{I}}}

{I}if (!fs::is_directory(root)) {{
{II}throw std::runtime_error(
{III}aas::common::Concat(
{IIII}"The path that you specified as a root directory which you wanted to "
{IIII}"list for subdirectories is not a directory: ",
{IIII}root.string()
{III})
{II});
{I}}}

{I}std::deque<fs::path> result;

{I}for (
{II}const fs::directory_entry& entry
{II}: fs::directory_iterator(root)
{II}) {{
{II}if (fs::is_directory(entry.path())) {{
{III}result.push_back(entry.path());
{II}}}
{I}}}

{I}std::sort(result.begin(), result.end());
{I}return result;
}}"""
        ),
        Stripped(
            f"""\
void AssertContentEqualsExpectedOrRecord(
{I}const std::string& content,
{I}const std::filesystem::path& path
) {{
{I}const bool record_mode = DetermineRecordMode();
{I}if (record_mode) {{
{II}fs::create_directories(path.parent_path());

{II}MustWriteString(path, content);
{I}}} else {{
{II}INFO(
{III}aas::common::Concat(
{IIII}"Expected to compare the content against "
{IIII}"the recorded file: ",
{IIII}path.string(),
{IIII}", but the file does not exist. If you want to re-record, "
{IIII}"please set the environment variable {environment_variable}_TEST_RECORD_MODE "
{IIII}"to 'ON'"
{III})
{II})
{II}REQUIRE(std::filesystem::exists(path));

{II}const std::string expected = MustReadString(path);
{II}INFO(
{III}aas::common::Concat(
{IIII}"Got unexpected content, which should have been equal "
{IIII}"to the content of the file: ",
{IIII}path.string()
{III})
{II})
{II}REQUIRE(content == expected);
{I}}}
}}"""
        ),
        Stripped(
            """\
}  // namespace common
}  // namespace test"""
        ),
        cpp_common.WARNING,
    ]

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(block)

    writer.write("\n")

    return writer.getvalue()


assert generate_header.__doc__ is not None
cpp_common.assert_module_docstring_and_generate_header_consistent(
    module_doc=__doc__, generate_header_doc=generate_header.__doc__
)

assert generate_implementation.__doc__ is not None
cpp_common.assert_module_docstring_and_generate_implementation_consistent(
    module_doc=__doc__, generate_implementation_doc=generate_implementation.__doc__
)



"""Generate the unit tests for Base64 functions."""

import io
from typing import List

from icontract import ensure

from aas_core_codegen.common import Stripped
from aas_core_codegen.typescript.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
    INDENT4 as IIII,
)


# fmt: off
@ensure(
    lambda result: result.endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate() -> str:
    """Generate the unit tests for Base64 functions."""
    blocks = [
        Stripped(
            """\
/**
 * Test base64 encoding and decoding.
 */"""
        ),
        Stripped(
            """\
import * as AasCommon from "../src/common";"""
        ),
        Stripped(
            f"""\
function testEncodeDecode(text: string, expectedEncoded: string): void {{
{I}const bytes = Uint8Array.from(text.split("").map((c) => c.charCodeAt(0)));

{I}const encoded = AasCommon.base64Encode(bytes);
{I}expect(encoded).toEqual(expectedEncoded);

{I}const decodedOrError: AasCommon.Either<Uint8Array, string> =
{II}AasCommon.base64Decode(encoded);

{I}expect(decodedOrError.error).toBeNull();
{I}expect(decodedOrError.mustValue()).toEqual(bytes);
}}"""
        ),
        Stripped(
            """\
// NOTE (mristin):
// The following tests come from:
// https://www.rfc-editor.org/rfc/rfc4648#section-10"""
        ),
        Stripped(
            f"""\
test("empty", () => {{
{I}testEncodeDecode("", "");
}});"""
        ),
        Stripped(
            f"""\
test("Uint8Array from 'f' is encoded 'Zg=='", () => {{
{I}testEncodeDecode("f", "Zg==");
}});"""
        ),
        Stripped(
            f"""\
test("Uint8Array from 'fo' is encoded 'Zm8='", () => {{
{I}testEncodeDecode("fo", "Zm8=");
}});"""
        ),
        Stripped(
            f"""\
test("Uint8Array from 'foo' is encoded 'Zm9v'", () => {{
{I}testEncodeDecode("foo", "Zm9v");
}});"""
        ),
        Stripped(
            f"""\
test("Uint8Array from 'foob' is encoded 'Zm9vYg=='", () => {{
{I}testEncodeDecode("foob", "Zm9vYg==");
}});"""
        ),
        Stripped(
            f"""\
test("Uint8Array from 'fooba' is encoded 'Zm9vYmE='", () => {{
{I}testEncodeDecode("fooba", "Zm9vYmE=");
}});"""
        ),
        Stripped(
            f"""\
test("Uint8Array from 'foobar' is encoded 'Zm9vYmFy'", () => {{
{I}testEncodeDecode("foobar", "Zm9vYmFy");
}});"""
        ),
        Stripped(
            f"""\
test("unexpected padding in the middle", () => {{
{I}const encoded = "Zm9vYmFy";

{I}for (let i = 0; i < encoded.length - 1; i++) {{
{II}const badEncoded =
{III}encoded.substring(0, i) + "=" + encoded.substring(i + 1, encoded.length);

{II}// Test the test
{II}expect(badEncoded.length).toEqual(encoded.length);

{II}const decodedOrError = AasCommon.base64Decode(badEncoded);
{II}expect(decodedOrError.error).toEqual(
{III}"Expected a valid character from " +
{IIII}"base64-encoded string, but got at " +
{IIII}`index ${{i}}: = (code: 61)`
{II});
{I}}}
}});"""
        ),
        # pylint: disable=line-too-long
        Stripped(
            f"""\
test("Bytes representing an invalid UTF-8 sequence encode OK", () => {{
{I}// NOTE (mristin):
{I}// We simply add a test to make sure our implementation does not suffer from
{I}// issues raised in:
{I}// https://stackoverflow.com/questions/30106476/using-javascripts-atob-to-decode-base64-doesnt-properly-decode-utf-8-strings

{I}// From:
{I}// https://stackoverflow.com/questions/1301402/example-invalid-utf8-string
{I}const bytes = new Uint8Array([0xc3, 0x28]);

{I}// NOTE (mristin):
{I}// This encoding was obtained by Python `base64` module.
{I}const expectedEncoded = "wyg=";

{I}const encoded = AasCommon.base64Encode(bytes);
{I}expect(encoded).toEqual(expectedEncoded);

{I}const decodedOrError: AasCommon.Either<Uint8Array, string> =
{II}AasCommon.base64Decode(encoded);

{I}expect(decodedOrError.error).toBeNull();
{I}expect(decodedOrError.mustValue()).toEqual(bytes);
}});"""
        ),
        Stripped(
            """\
// NOTE (mristin):
// The following checks come from: https://eprint.iacr.org/2022/361.pdf"""
        ),
        Stripped(
            f"""\
test("'Hello' as 'SGVsbG8=' OK", () => {{
{I}testEncodeDecode("Hello", "SGVsbG8=");
}});"""
        ),
        Stripped(
            f"""\
test("our implementation suffers from padding inconsistency of 'Hello' as 'SGVsbG9='", () => {{
{I}// NOTE (mristin):
{I}// This is not a test case, but we merely document that there is a possible
{I}// attack vector where different strings encode the *same* byte sequence.

{I}const badEncoded = "SGVsbG9=";

{I}const decodedOrError = AasCommon.base64Decode(badEncoded);
{I}expect(decodedOrError.mustValue()).toEqual(
{II}new Uint8Array(
{III}// "Hello"
{III}[72, 101, 108, 108, 111]
{II})
{I});
{I}expect(decodedOrError.error).toBeNull();
}});"""
        ),
        Stripped(
            f"""\
test("table from rickkas7/Base64RK", () => {{
{I}// NOTE (mristin):
{I}// The following tests have been scraped from:
{I}// https://github.com/rickkas7/Base64RK/blob/183d20e62b96ef9b0230c80c7a172715ca6f661e/test/unit-test/unit-test.cpp

{I}const arraysExpectedEncodeds: Array<[Array<number>, string]> = [
{II}[[28], "HA=="],
{II}[[174, 208], "rtA="],
{II}[[50, 224, 208], "MuDQ"],
{II}[[45, 18, 235, 6], "LRLrBg=="],
{II}[[117, 199, 153, 221, 160], "dceZ3aA="],
{II}[[141, 12, 188, 48, 173, 45], "jQy8MK0t"],
{II}[[69, 143, 148, 125, 213, 3, 148], "RY+UfdUDlA=="],
{II}[[251, 162, 166, 60, 129, 131, 46, 13], "+6KmPIGDLg0="],
{II}[[207, 40, 204, 254, 77, 71, 61, 8, 128], "zyjM/k1HPQiA"],
{II}[[156, 150, 213, 174, 117, 59, 70, 220, 210, 63], "nJbVrnU7RtzSPw=="],
{II}[[192, 54, 173, 19, 126, 84, 26, 250, 167, 5, 50], "wDatE35UGvqnBTI="],
{II}[[20, 1, 15, 77, 224, 141, 153, 47, 186, 170, 200, 227], "FAEPTeCNmS+6qsjj"],
{II}[
{III}[85, 166, 130, 169, 193, 117, 194, 81, 238, 64, 180, 228, 127],
{III}"VaaCqcF1wlHuQLTkfw=="
{II}],
{II}[
{III}[65, 130, 145, 17, 247, 99, 158, 36, 117, 8, 192, 1, 193, 234],
{III}"QYKREfdjniR1CMABweo="
{II}],
{II}[
{III}[9, 55, 253, 179, 203, 93, 237, 58, 252, 200, 168, 145, 115, 53, 31],
{III}"CTf9s8td7Tr8yKiRczUf"
{II}],
{II}[
{III}[
{IIII}254, 76, 140, 77, 16, 52, 119, 229, 186, 225, 109, 181, 134, 34, 229, 153, 235,
{IIII}1, 16, 126, 17, 14, 13, 12, 138, 248, 230, 60, 73, 128, 116, 204, 207
{III}],
{III}"/kyMTRA0d+W64W21hiLlmesBEH4RDg0MivjmPEmAdMzP"
{II}],
{II}[
{III}[
{IIII}155, 159, 208, 168, 221, 216, 113, 8, 58, 136, 77, 93, 86, 50, 80, 91, 14, 102,
{IIII}10, 54, 2, 158, 195, 75, 40, 245, 84, 177, 116, 54, 235, 217, 124, 232, 205,
{IIII}113, 111, 33, 231, 251, 246, 136, 41, 167, 201, 234, 20, 83, 225, 67, 131, 210,
{IIII}97, 194, 1, 252, 158, 253, 84, 135, 119, 1, 254, 11, 182, 208
{III}],
{III}"m5/QqN3YcQg6iE1dVjJQWw5mCjYCnsNLKPVUsXQ269l86M1xbyHn+/aIKafJ6hRT4UOD0mHCAfye/VSHdwH+C7bQ"
{II}],
{II}[
{III}[
{IIII}101, 43, 33, 221, 108, 128, 221, 221, 52, 11, 40, 166, 120, 198, 79, 0, 141,
{IIII}127, 183, 226, 197, 232, 52, 29, 136, 1, 16, 27, 216, 71, 149, 251, 131, 94, 22,
{IIII}118, 25, 255, 214, 119, 91, 182, 203, 228, 67, 22, 88, 180, 161, 42, 182, 125,
{IIII}183, 239, 186, 216, 23, 11, 17, 144, 124, 201, 177, 3, 83, 68, 131, 249, 52,
{IIII}189, 2, 28, 74, 172, 214, 117, 126, 249, 239, 90, 22, 204, 161, 243, 232, 166,
{IIII}103, 150, 20, 240, 60, 93, 187, 67, 57, 231, 1, 102, 221, 194, 76, 162, 8, 89,
{IIII}72, 150, 146, 247, 181, 134, 242, 205, 179, 19, 155, 68, 94, 82, 214, 56, 1,
{IIII}111, 109, 78, 155, 216, 73, 7, 80, 243
{III}],
{III}"ZSsh3WyA3d00CyimeMZPAI1/t+LF6DQdiAEQG9hHlfuDXhZ2Gf/Wd1u2y+RDFli0oSq2fbfvutgXCxGQfMmxA1NEg/k0vQIcSqzWdX7571oWzKHz6KZnlhTwPF27QznnAWbdwkyiCFlIlpL3tYbyzbMTm0ReUtY4AW9tTpvYSQdQ8w=="
{II}],
{II}[
{III}[
{IIII}177, 224, 149, 210, 54, 194, 123, 141, 157, 224, 99, 1, 189, 114, 37, 230, 250,
{IIII}149, 161, 158, 138, 226, 5, 18, 131, 169, 202, 130, 46, 109, 102, 35, 113, 199,
{IIII}109, 178, 98, 137, 249, 69, 227, 130, 232, 246, 105, 101, 97, 234, 120, 151, 31,
{IIII}41, 90, 20, 15, 246, 90, 214, 39, 139, 86, 52, 78, 174, 35, 95, 63, 226, 241,
{IIII}127, 164, 219, 174, 140, 255, 17, 120, 118, 97, 131, 133, 35, 231, 117, 215, 74,
{IIII}14, 110, 70, 42, 110, 7, 37, 234, 139, 210, 224, 36, 71, 227, 68, 155, 236, 191,
{IIII}61, 83, 11, 189, 88, 60, 208, 3, 109, 121, 32, 71, 169, 58, 125, 112, 116, 16,
{IIII}153, 200, 74, 242, 89, 141, 143, 25, 211, 92, 179, 227, 223, 183, 219, 18, 54,
{IIII}208, 80, 253, 11, 76, 69, 150, 4, 48, 238, 73, 165, 25, 133, 200, 80, 207, 84,
{IIII}182, 64, 251, 156, 76, 255, 112, 227, 235, 154, 29, 191, 86, 217, 118, 219, 51,
{IIII}38, 173, 6, 111, 14, 55, 195, 13, 45, 236, 66, 180, 39, 109, 110, 7, 47, 61,
{IIII}103, 100, 117, 55, 89, 35, 64, 106, 106, 42, 238, 37, 70, 20, 8, 128, 34, 131,
{IIII}56, 120, 103, 125, 149, 182, 130, 217, 38, 102, 24, 170, 111, 47, 113, 79, 142,
{IIII}4, 167, 80, 12, 174, 102, 74, 129, 109, 204, 12, 252, 200, 95, 97, 51, 215, 169,
{IIII}148, 191, 113, 155, 228, 145, 235, 35, 252, 53, 177
{III}],
{III}"seCV0jbCe42d4GMBvXIl5vqVoZ6K4gUSg6nKgi5tZiNxx22yYon5ReOC6PZpZWHqeJcfKVoUD/Za1ieLVjROriNfP+Lxf6Tbroz/EXh2YYOFI+d110oObkYqbgcl6ovS4CRH40Sb7L89Uwu9WDzQA215IEepOn1wdBCZyEryWY2PGdNcs+Pft9sSNtBQ/QtMRZYEMO5JpRmFyFDPVLZA+5xM/3Dj65odv1bZdtszJq0Gbw43ww0t7EK0J21uBy89Z2R1N1kjQGpqKu4lRhQIgCKDOHhnfZW2gtkmZhiqby9xT44Ep1AMrmZKgW3MDPzIX2Ez16mUv3Gb5JHrI/w1sQ=="
{II}]
{I}];

{I}for (const [array, expectedEncoded] of arraysExpectedEncodeds) {{
{II}const bytes = new Uint8Array(array);

{II}const encoded = AasCommon.base64Encode(bytes);
{II}expect(encoded).toEqual(expectedEncoded);

{II}const decodedOrError: AasCommon.Either<Uint8Array, string> =
{III}AasCommon.base64Decode(encoded);

{II}expect(decodedOrError.error).toBeNull();
{II}expect(decodedOrError.mustValue()).toEqual(bytes);
{I}}}
}});"""
        ),
        # pylint: enable=line-too-long
    ]  # type: List[Stripped]
    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(block)

    writer.write("\n")

    return writer.getvalue()


assert generate.__doc__ is not None
assert generate.__doc__.strip().startswith(__doc__.strip())



"""Generate code to test the XML de/serialization of enumerations."""

import io
from typing import List, Optional

from icontract import ensure

from aas_core_codegen import intermediate, naming
from aas_core_codegen.common import Stripped, Identifier
from aas_core_codegen.typescript import (
    common as typescript_common,
    naming as typescript_naming,
)
from aas_core_codegen.typescript.common import (
    INDENT as I,
    INDENT2 as II,
)


# fmt: off
@ensure(
    lambda result: result.endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate(symbol_table: intermediate.SymbolTable) -> str:
    """Generate code to test the XML de/serialization of enumerations."""
    blocks = [
        Stripped(
            """\
/**
 * Test XML de/serialization of enumeration literals.
 */"""
        ),
        typescript_common.WARNING,
        Stripped(
            """\
import * as AasXmlization from "../src/xmlization";
import * as AasTypes from "../src/types";

import * as TestCommonXmlization from "./commonXmlization";"""
        ),
    ]  # type: List[Stripped]

    for enumeration in symbol_table.enumerations:
        enum_name_typescript = typescript_naming.enum_name(enumeration.name)

        carrier_cls = None  # type: Optional[intermediate.ConcreteClass]
        carrier_prop = None  # type: Optional[intermediate.Property]
        for cls in symbol_table.concrete_classes:
            for prop in cls.properties:
                type_anno = intermediate.beneath_optional(prop.type_annotation)
                if not isinstance(type_anno, intermediate.OurTypeAnnotation):
                    continue

                if type_anno.our_type is enumeration:
                    carrier_cls = cls
                    carrier_prop = prop
                    break

            if carrier_cls is not None:
                break

        if carrier_cls is None or carrier_prop is None:
            raise AssertionError(
                f"Expected an enum carrier property for XML round-trip tests, "
                f"but found none for: {enumeration.name}"
            )

        prop_name_typescript = typescript_naming.property_name(carrier_prop.name)
        load_maximal_name = typescript_naming.function_name(
            Identifier(f"load_maximal_{carrier_cls.name}")
        )
        as_function = typescript_naming.function_name(
            Identifier(f"as_{carrier_cls.name}")
        )

        blocks.append(
            Stripped(
                f"""\
test("{enum_name_typescript} XML round-trip OK", () => {{
{I}const instance = TestCommonXmlization.{load_maximal_name}();

{I}const xmlText = AasXmlization.toXmlString(instance);
{I}const anotherOrError = AasXmlization.fromXmlString(xmlText);
{I}expect(anotherOrError.error).toBeNull();

{I}const casted = AasTypes.{as_function}(anotherOrError.mustValue());
{I}expect(casted).not.toBeNull();
{I}expect(casted.{prop_name_typescript}).toStrictEqual(
{II}instance.{prop_name_typescript}
{I});
}});"""
            )
        )

        literal_value_set = set(literal.value for literal in enumeration.literals)
        invalid_literal_value = "invalid-literal"
        while invalid_literal_value in literal_value_set:
            invalid_literal_value = f"very-{invalid_literal_value}"

        prop_xml_name_literal = typescript_common.string_literal(
            naming.xml_property(carrier_prop.name)
        )
        invalid_literal_literal = typescript_common.string_literal(
            invalid_literal_value
        )

        blocks.append(
            Stripped(
                f"""\
test("{enum_name_typescript} XML deserialization fail", () => {{
{I}const instance = TestCommonXmlization.{load_maximal_name}();
{I}const xmlText = AasXmlization.toXmlString(instance);

{I}const regex = new RegExp(
{II}`(<${{{prop_xml_name_literal}}}>)([^<]*)(</${{{prop_xml_name_literal}}}>)`
{I});
{I}const brokenXmlText = xmlText.replace(
{II}regex,
{II}`$1${{{invalid_literal_literal}}}$3`
{I});

{I}expect(brokenXmlText).not.toStrictEqual(xmlText);

{I}const anotherOrError = AasXmlization.fromXmlString(brokenXmlText);
{I}expect(anotherOrError.error).not.toBeNull();
}});"""
            )
        )

    blocks.append(typescript_common.WARNING)

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(block)

    writer.write("\n")

    return writer.getvalue()


assert generate.__doc__ is not None
assert generate.__doc__.strip().startswith(__doc__.strip())



"""Generate code to test the JSON de/serialization of enumerations."""

import io
from typing import List

from icontract import ensure

from aas_core_codegen import intermediate
from aas_core_codegen.common import Stripped, Identifier
from aas_core_codegen.typescript import (
    common as typescript_common,
    naming as typescript_naming,
)
from aas_core_codegen.typescript.common import (
    INDENT as I,
    INDENT2 as II,
)


# fmt: off
@ensure(
    lambda result: result.endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate(symbol_table: intermediate.SymbolTable) -> str:
    """Generate code to test the JSON de/serialization of enumerations."""
    blocks = [
        Stripped(
            """\
/**
 * Test JSON de/serialization of enumeration literals.
 */"""
        ),
        typescript_common.WARNING,
        Stripped(
            """\
import * as AasJsonization from "../src/jsonization";
import * as AasTypes from "../src/types";"""
        ),
    ]  # type: List[Stripped]

    for enumeration in symbol_table.enumerations:
        enum_name_typescript = typescript_naming.enum_name(enumeration.name)

        assert (
            len(enumeration.literals) > 0
        ), f"Unexpected enumeration without literals: {enumeration.name}"

        literal = enumeration.literals[0]
        literal_name_typescript = typescript_naming.enum_literal_name(literal.name)

        deserialization_function = typescript_naming.function_name(
            Identifier(f"{enumeration.name}_from_jsonable")
        )

        blocks.append(
            Stripped(
                f"""\
test("{enum_name_typescript} round-trip OK", () => {{
{I}const jsonable = {typescript_common.string_literal(literal.value)};

{I}const literalOrError = AasJsonization.{deserialization_function}(
{II}jsonable
{I});

{I}expect(literalOrError.error).toBeNull();
{I}const literal = literalOrError.mustValue();

{I}expect(literal).toStrictEqual(
{II}AasTypes.{enum_name_typescript}.{literal_name_typescript}
{I});
}});"""
            )
        )

        literal_value_set = set(literal.value for literal in enumeration.literals)
        invalid_literal_value = "invalid-literal"
        while invalid_literal_value in literal_value_set:
            invalid_literal_value = f"very-{invalid_literal_value}"

        expected_message = (
            f"Not a valid string representation of a literal "
            f"of {enum_name_typescript}: {invalid_literal_value}"
        )

        blocks.append(
            Stripped(
                f"""\
test("{enum_name_typescript} deserialization fail", () => {{
{I}const jsonable = {typescript_common.string_literal(invalid_literal_value)};

{I}const literalOrError = AasJsonization.{deserialization_function}(
{II}jsonable
{I});

{I}expect(literalOrError.error.message).toStrictEqual(
{II}{typescript_common.string_literal(expected_message)}
{I});
}});"""
            )
        )

    blocks.append(typescript_common.WARNING)

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(block)

    writer.write("\n")

    return writer.getvalue()


assert generate.__doc__ is not None
assert generate.__doc__.strip().startswith(__doc__.strip())



"""Generate code to test the ``typesMatch`` function."""

import io
from typing import List

from icontract import ensure

from aas_core_codegen import intermediate
from aas_core_codegen.common import Stripped, Identifier, indent_but_first_line
from aas_core_codegen.typescript import (
    common as typescript_common,
    naming as typescript_naming,
)
from aas_core_codegen.typescript.common import (
    INDENT as I,
    INDENT2 as II,
)


# fmt: off
@ensure(
    lambda result: result.endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate(symbol_table: intermediate.SymbolTable) -> str:
    """Generate code to test the ``typesMatch`` function."""
    blocks = [
        Stripped(
            """\
/**
 * Test `typesMatch`.
 */"""
        ),
        typescript_common.WARNING,
        Stripped(
            """\
import * as AasTypes from "../src/types";
import * as TestCommonJsonization from "./commonJsonization";"""
        ),
    ]  # type: List[Stripped]

    for concrete_cls in symbol_table.concrete_classes:
        load_minimal_name = typescript_naming.function_name(
            Identifier(f"load_minimal_{concrete_cls.name}")
        )

        instance_var = typescript_naming.constant_name(
            Identifier(f"the_{concrete_cls.name}")
        )

        blocks.append(
            Stripped(
                f"""\
const {instance_var} = TestCommonJsonization.{load_minimal_name}();"""
            )
        )

    for concrete_cls in symbol_table.concrete_classes:
        body_blocks = []  # type: List[Stripped]

        that_var = typescript_naming.constant_name(
            Identifier(f"the_{concrete_cls.name}")
        )

        for other_cls in symbol_table.concrete_classes:
            other_var = typescript_naming.constant_name(
                Identifier(f"the_{other_cls.name}")
            )

            if other_cls.is_subclass_of(concrete_cls):
                body_blocks.append(
                    Stripped(
                        f"""\
expect(
{I}AasTypes.typesMatch(
{II}{that_var},
{II}{other_var}
{I})
).toStrictEqual(true);"""
                    )
                )
            else:
                body_blocks.append(
                    Stripped(
                        f"""\
expect(
{I}AasTypes.typesMatch(
{II}{that_var},
{II}{other_var}
{I})
).toStrictEqual(false);"""
                    )
                )

        body = "\n\n".join(body_blocks)

        cls_name_typescript = typescript_naming.class_name(concrete_cls.name)
        blocks.append(
            Stripped(
                f"""\
test("type matches for {cls_name_typescript}", () => {{
{I}{indent_but_first_line(body, I)}
}});"""
            )
        )

    blocks.append(typescript_common.WARNING)

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(block)

    writer.write("\n")

    return writer.getvalue()


assert generate.__doc__ is not None
assert generate.__doc__.strip().startswith(__doc__.strip())



"""Generate the unit test code for TypeScript."""

from aas_core_codegen.typescript.tests import (
    _generate_common,
    _generate_common_base64_spec,
    _generate_common_base64url_spec,
    _generate_common_jsonization,
    _generate_common_xmlization,
    _generate_jsonization_concrete_classes_spec,
    _generate_jsonization_enums_spec,
    _generate_jsonization_interfaces_spec,
    _generate_xmlization_concrete_classes_spec,
    _generate_xmlization_enums_spec,
    _generate_xmlization_interfaces_spec,
    _generate_types_casts_spec,
    _generate_types_descend_and_pass_through_visitor_spec,
    _generate_types_descend_once_spec,
    _generate_types_model_type_spec,
    _generate_types_over_enum_spec,
    _generate_types_over_x_or_empty_spec,
    _generate_types_type_matches_spec,
    _generate_types_x_or_default_spec,
)

generate_common = _generate_common.generate
generate_common_base64_spec = _generate_common_base64_spec.generate
generate_common_base64url_spec = _generate_common_base64url_spec.generate
generate_common_jsonization = _generate_common_jsonization.generate
generate_common_xmlization = _generate_common_xmlization.generate
generate_jsonization_concrete_classes_spec = (
    _generate_jsonization_concrete_classes_spec.generate
)
generate_jsonization_enums_spec = _generate_jsonization_enums_spec.generate
generate_jsonization_interfaces_spec = _generate_jsonization_interfaces_spec.generate
generate_xmlization_concrete_classes_spec = (
    _generate_xmlization_concrete_classes_spec.generate
)
generate_xmlization_enums_spec = _generate_xmlization_enums_spec.generate
generate_xmlization_interfaces_spec = _generate_xmlization_interfaces_spec.generate
generate_types_casts_spec = _generate_types_casts_spec.generate
generate_types_descend_and_pass_through_visitor_spec = (
    _generate_types_descend_and_pass_through_visitor_spec.generate
)
generate_types_descend_once_spec = _generate_types_descend_once_spec.generate
generate_types_model_type_spec = _generate_types_model_type_spec.generate
generate_types_over_enum_spec = _generate_types_over_enum_spec.generate
generate_types_over_x_or_empty_spec = _generate_types_over_x_or_empty_spec.generate
generate_types_type_matches_spec = _generate_types_type_matches_spec.generate
generate_types_x_or_default_spec = _generate_types_x_or_default_spec.generate



"""Generate code to test the ``OverXOrEmpty`` methods."""

import io
from typing import List

from icontract import ensure

from aas_core_codegen import intermediate
from aas_core_codegen.common import Stripped, Identifier
from aas_core_codegen.typescript import (
    common as typescript_common,
    naming as typescript_naming,
)
from aas_core_codegen.typescript.common import (
    INDENT as I,
    INDENT2 as II,
)


# fmt: off
@ensure(
    lambda result: result.endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate(symbol_table: intermediate.SymbolTable) -> str:
    """Generate code to test the ``OverXOrEmpty`` methods."""
    blocks = [
        Stripped(
            """\
/**
 * Test `over*OrEmpty` methods.
 */"""
        ),
        typescript_common.WARNING,
        Stripped(
            """\
import * as TestCommonJsonization from "./commonJsonization";"""
        ),
    ]  # type: List[Stripped]

    for concrete_cls in symbol_table.concrete_classes:
        cls_name_typescript = typescript_naming.class_name(concrete_cls.name)

        for prop in concrete_cls.properties:
            method_name_typescript = typescript_naming.method_name(
                Identifier(f"Over_{prop.name}_or_empty")
            )

            prop_name_typescript = typescript_naming.property_name(prop.name)

            if isinstance(
                prop.type_annotation, intermediate.OptionalTypeAnnotation
            ) and isinstance(
                prop.type_annotation.value, intermediate.ListTypeAnnotation
            ):
                load_maximal_name = typescript_naming.function_name(
                    Identifier(f"load_maximal_{concrete_cls.name}")
                )

                # noinspection SpellCheckingInspection
                blocks.append(
                    Stripped(
                        f"""\
test("{cls_name_typescript}.{method_name_typescript} on maximal", () => {{
{I}const instance = TestCommonJsonization.{load_maximal_name}();

{I}let count = 0;
{I}for (const _ of instance.{method_name_typescript}()) {{
{II}count++;
{I}}}

{I}expect(count).toStrictEqual(
{II}instance.{prop_name_typescript}?.length ?? 0,
{I});
}});"""
                    )
                )

                load_minimal_name = typescript_naming.function_name(
                    Identifier(f"load_minimal_{concrete_cls.name}")
                )

                blocks.append(
                    Stripped(
                        f"""\
test("{cls_name_typescript}.{method_name_typescript} on minimal", () => {{
{I}const instance = TestCommonJsonization.{load_minimal_name}();

{I}let count = 0;
{I}for (const _ of instance.{method_name_typescript}()) {{
{II}count++;
{I}}}

{I}expect(count).toStrictEqual(
{II}instance.{prop_name_typescript}?.length ?? 0,
{I});
}});"""
                    )
                )

    blocks.append(typescript_common.WARNING)

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(block)

    writer.write("\n")

    return writer.getvalue()


assert generate.__doc__ is not None
assert generate.__doc__.strip().startswith(__doc__.strip())



"""Generate code to load minimal and maximal examples."""


import io
from typing import List

from icontract import ensure

from aas_core_codegen import intermediate, naming
from aas_core_codegen.common import (
    Stripped,
    Identifier,
)
from aas_core_codegen.cpp import common as cpp_common, naming as cpp_naming
from aas_core_codegen.cpp.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
    INDENT4 as IIII,
)


def _generate_load_min_max_definitions(
    cls: intermediate.ConcreteClass, library_namespace: Stripped
) -> List[Stripped]:
    """Generate the def. of loading minimal and maximal example, respectively."""
    interface_name = cpp_naming.interface_name(cls.name)

    load_min = cpp_naming.function_name(Identifier(f"load_min_{cls.name}"))
    load_max = cpp_naming.function_name(Identifier(f"load_max_{cls.name}"))

    return [
        Stripped(
            f"""\
/**
 * Load a minimal example of {interface_name} by
 * de-serializing it from an XML file.
 */
std::shared_ptr<
{I}{library_namespace}::types::{interface_name}
> {load_min}();"""
        ),
        Stripped(
            f"""\
/**
 * Load a maximal example of {interface_name} by
 * de-serializing it from an XML file.
 */
std::shared_ptr<
{I}{library_namespace}::types::{interface_name}
> {load_max}();"""
        ),
    ]


# fmt: off
@ensure(
    lambda result:
    result.endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate_header(
    symbol_table: intermediate.SymbolTable, library_namespace: Stripped
) -> str:
    """Generate header to load minimal and maximal examples."""
    include_guard_var = cpp_common.include_guard_var(Stripped("test::common::examples"))

    include_prefix_path = cpp_common.generate_include_prefix_path(library_namespace)

    blocks = [
        cpp_common.WARNING,
        # NOTE (mristin):
        # We de-serialize from an XML file instead of JSON to avoid problems
        # with missing `modelType` JSON properties for some classes such as
        # `Abstract_lang_string`.
        Stripped(
            """\
/**
* Load minimal and maximal examples from an XML file.
*/"""
        ),
        Stripped(
            f"""\
#ifndef {include_guard_var}
#define {include_guard_var}

#include <{include_prefix_path}/types.hpp>

namespace test {{
namespace common {{
namespace examples {{"""
        ),
    ]

    for concrete_cls in symbol_table.concrete_classes:
        blocks.extend(
            _generate_load_min_max_definitions(
                cls=concrete_cls, library_namespace=library_namespace
            )
        )

    blocks.extend(
        [
            Stripped(
                f"""\
}}  // namespace examples
}}  // namespace common
}}  // namespace test

#endif // {include_guard_var}"""
            ),
            cpp_common.WARNING,
        ]
    )

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(block)

    writer.write("\n")

    return writer.getvalue()


def _generate_static_type_name(cls: intermediate.ConcreteClass) -> Stripped:
    """Generate the specialization of a templated struct to retrieve the type name."""
    interface_name = cpp_naming.interface_name(cls.name)

    return Stripped(
        f"""\
template<>
struct StaticTypeName<
{I}aas::types::{interface_name}
> {{
  static const char* name;
}};
const char* StaticTypeName<
{I}aas::types::{interface_name}
>::name = "{interface_name}";"""
    )


def _generate_load_min_max_implementations(
    cls: intermediate.ConcreteClass,
) -> List[Stripped]:
    """Generate the impl. of loading minimal and maximal example, respectively."""
    interface_name = cpp_naming.interface_name(cls.name)

    load = cpp_naming.function_name(Identifier(f"load_{cls.name}"))
    load_min = cpp_naming.function_name(Identifier(f"load_min_{cls.name}"))
    load_max = cpp_naming.function_name(Identifier(f"load_max_{cls.name}"))

    xml_class_name = naming.xml_class_name(cls.name)

    return [
        Stripped(
            f"""\
std::shared_ptr<
{I}aas::types::{interface_name}
> {load}(
{I}const std::filesystem::path& path
) {{
{I}

{I}std::shared_ptr<
{II}aas::types::IClass
{I}> abstract = test::common::xmlization::MustReadInstance(
{II}path
{I});

{I}std::shared_ptr<
{II}aas::types::{interface_name}
{I}> instance(
{II}std::dynamic_pointer_cast<
{III}aas::types::{interface_name}
{II}>(
{III}abstract
{II})
{I});

{I}if (instance == nullptr) {{
{II}throw std::runtime_error(
{III}aas::common::Concat(
{IIII}"Failed to cast the instance to {interface_name} from ",
{IIII}path.string()
{III})
{II});
{I}}}

{I}return instance;
}}"""
        ),
        Stripped(
            f"""\
std::shared_ptr<
{I}aas::types::{interface_name}
> {load_min}() {{
{I}const std::filesystem::path path(
{II}test::common::DetermineTestDataDir()
{III}/ "Xml"
{III}/ "Expected"
{III}/ {cpp_common.string_literal(xml_class_name)}
{III}/ "minimal.xml"
{I});

{I}return {load}(
{II}path
{I});
}}"""
        ),
        Stripped(
            f"""\
std::shared_ptr<
{I}aas::types::{interface_name}
> {load_max}() {{
{I}const std::filesystem::path path(
{II}test::common::DetermineTestDataDir()
{III}/ "Xml"
{III}/ "Expected"
{III}/ {cpp_common.string_literal(xml_class_name)}
{III}/ "maximal.xml"
{I});

{I}return {load}(
{II}path
{I});
}}"""
        ),
    ]


# fmt: off
@ensure(
    lambda result:
    result.endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate_implementation(
    symbol_table: intermediate.SymbolTable, library_namespace: Stripped
) -> str:
    """Generate implementation to load minimal and maximal examples."""
    include_prefix_path = cpp_common.generate_include_prefix_path(library_namespace)

    blocks = [
        cpp_common.WARNING,
        Stripped(
            f"""\
#include "./common.hpp"
#include "./common_examples.hpp"
#include "./common_xmlization.hpp"

#include <{include_prefix_path}/iteration.hpp>
#include <{include_prefix_path}/stringification.hpp>

#include <filesystem>

namespace aas = {library_namespace};

namespace test {{
namespace common {{
namespace examples {{"""
        ),
        Stripped(
            """\
template<typename T>
struct StaticTypeName;"""
        ),
    ]

    for concrete_cls in symbol_table.concrete_classes:
        blocks.append(_generate_static_type_name(concrete_cls))

    for concrete_cls in symbol_table.concrete_classes:
        blocks.extend(_generate_load_min_max_implementations(cls=concrete_cls))

    blocks.extend(
        [
            Stripped(
                """\
}  // namespace examples
}  // namespace common
}  // namespace test"""
            ),
            cpp_common.WARNING,
        ]
    )

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(block)

    writer.write("\n")

    return writer.getvalue()


assert generate_header.__doc__ is not None
cpp_common.assert_module_docstring_and_generate_header_consistent(
    module_doc=__doc__, generate_header_doc=generate_header.__doc__
)

assert generate_implementation.__doc__ is not None
cpp_common.assert_module_docstring_and_generate_implementation_consistent(
    module_doc=__doc__, generate_implementation_doc=generate_implementation.__doc__
)



"""Generate code to test the XML de/serialization of concrete classes."""

import io
from typing import List

from icontract import ensure

from aas_core_codegen import intermediate, naming
from aas_core_codegen.common import (
    Stripped,
    Identifier,
)
from aas_core_codegen.golang import common as golang_common, naming as golang_naming
from aas_core_codegen.golang.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
    INDENT4 as IIII,
    INDENT5 as IIIII,
)


def _generate_for_cls(cls: intermediate.ConcreteClass) -> List[Stripped]:
    """Generate the tests for a self-contained class."""
    xml_class_name_literal = golang_common.string_literal(
        naming.xml_class_name(cls.name)
    )

    unmarshal_function = golang_naming.function_name(Identifier("unmarshal"))

    blocks = []  # type: List[Stripped]

    interface_name = golang_naming.interface_name(cls.name)

    test_name = golang_naming.function_name(
        Identifier(f"Test_{cls.name}_round_trip_OK")
    )

    blocks.append(
        Stripped(
            f"""\
func {test_name}(t *testing.T) {{
{I}pths := aastesting.FindFilesBySuffixRecursively(
{II}filepath.Join(
{III}aastesting.TestDataDir,
{III}"Xml",
{III}"Expected",
{III}{xml_class_name_literal},
{II}),
{II}".xml",
{I})
{I}sort.Strings(pths)

{I}for _, pth := range pths {{
{II}bb, err := os.ReadFile(pth)
{II}if err != nil {{
{III}t.Fatalf("Failed to read the file %s: %s", pth, err.Error())
{III}return
{II}}}
{II}text := string(bb)

{II}decoder := xml.NewDecoder(strings.NewReader(text))

{II}deserialized, deseriaErr := aasxmlization.{unmarshal_function}(decoder)
{II}ok := assertNoDeserializationError(t, deseriaErr, pth)
{II}if !ok {{
{III}return
{II}}}

{II}if _, ok := deserialized.(aastypes.{interface_name}); !ok {{
{III}t.Fatalf(
{IIII}"Expected an instance of {interface_name}, "+
{IIIII}"but got %T: %v",
{IIII}deserialized, deserialized,
{III})
{III}return
{II}}}

{II}buf := &bytes.Buffer{{}}
{II}encoder := xml.NewEncoder(buf)
{II}encoder.Indent("", "\\t")

{II}seriaErr := aasxmlization.Marshal(encoder, deserialized, true)
{II}ok = assertNoSerializationError(t, seriaErr, pth)
{II}if !ok {{
{III}return
{II}}}

{II}roundTrip := string(buf.Bytes())

{II}ok = assertSerializationEqualsDeserialization(
{III}t,
{III}text,
{III}roundTrip,
{III}pth,
{II})
{II}if !ok {{
{III}return
{II}}}
{I}}}
}}"""
        )
    )

    test_name = golang_naming.function_name(
        Identifier(f"Test_{cls.name}_deserialization_fail")
    )

    blocks.append(
        Stripped(
            f"""\
func {test_name}(t *testing.T) {{
{I}pattern := filepath.Join(
{II}aastesting.TestDataDir,
{II}"Xml",
{II}"Unexpected",
{II}"Unserializable",
{II}"*",  // This asterisk represents the cause.
{II}{xml_class_name_literal},
{I})

{I}causeDirs, err := filepath.Glob(pattern)
{I}if err != nil {{
{II}panic(
{III}fmt.Sprintf(
{IIII}"Failed to find cause directories matching %s: %s",
{IIII}pattern, err.Error(),
{III}),
{II})
{I}}}

{I}for _, causeDir := range causeDirs {{
{II}pths := aastesting.FindFilesBySuffixRecursively(
{III}causeDir,
{III}".xml",
{II})
{II}sort.Strings(pths)

{II}for _, pth := range pths {{
{III}relPth, err := filepath.Rel(aastesting.TestDataDir, pth)
{III}if err != nil {{
{IIII}panic(
{IIIII}fmt.Sprintf(
{IIIII}{I}"Failed to compute the relative path of %s to %s: %s",
{IIIII}{I}aastesting.TestDataDir, pth, err.Error(),
{IIIII}),
{IIII})
{III}}}

{III}expectedPth := filepath.Join(
{IIII}aastesting.TestDataDir,
{IIII}"DeserializationError",
{IIII}filepath.Dir(relPth),
{IIII}filepath.Base(relPth)+".error",
{III})

{III}bb, err := os.ReadFile(pth)
{III}if err != nil {{
{IIII}t.Fatalf("Failed to read the file %s: %s", pth, err.Error())
{IIII}return
{III}}}
{III}text := string(bb)

{III}decoder := xml.NewDecoder(strings.NewReader(text))

{III}_, deseriaErr := aasxmlization.Unmarshal(decoder)
{III}ok := assertIsDeserializationErrorAndEqualsExpectedOrRecord(
{IIII}t, deseriaErr, pth, expectedPth,
{III})
{III}if !ok {{
{IIII}return
{III}}}
{II}}}
{I}}}
}}"""
        )
    )

    return blocks


# fmt: off
@ensure(
    lambda result: result.endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate(symbol_table: intermediate.SymbolTable, repo_url: Stripped) -> str:
    """Generate code to test the XML de/serialization of concrete classes."""
    blocks = [
        Stripped("package xmlization_test"),
        golang_common.WARNING,
        Stripped(
            f"""\
import (
{I}"bytes"
{I}"path/filepath"
{I}"fmt"
{I}"os"
{I}"sort"
{I}"strings"
{I}"testing"
{I}"encoding/xml"
{I}aastesting "{repo_url}/aastesting"
{I}aastypes "{repo_url}/types"
{I}aasxmlization "{repo_url}/xmlization"
)"""
        ),
    ]  # type: List[Stripped]

    for concrete_cls in symbol_table.concrete_classes:
        blocks.extend(_generate_for_cls(cls=concrete_cls))

    blocks.append(golang_common.WARNING)

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(block)

    writer.write("\n")

    return writer.getvalue()


assert generate.__doc__ is not None
assert generate.__doc__.strip().startswith(__doc__.strip())



"""Generate code to load minimal and maximal examples."""

import io
from typing import List

from icontract import ensure

from aas_core_codegen import intermediate, naming
from aas_core_codegen.common import (
    Stripped,
    Identifier,
    indent_but_first_line,
)
from aas_core_codegen.golang import common as golang_common, naming as golang_naming
from aas_core_codegen.golang.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
)


# fmt: off
@ensure(
    lambda result: result.endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate(symbol_table: intermediate.SymbolTable, repo_url: Stripped) -> str:
    """Generate code to load minimal and maximal examples."""
    blocks = [
        Stripped("package aastesting"),
        golang_common.WARNING,
        Stripped(
            f"""\
import (
{I}"fmt"
{I}"path"
{I}aasjsonization "{repo_url}/jsonization"
{I}aastypes "{repo_url}/types"
)"""
        ),
    ]  # type: List[Stripped]

    for concrete_cls in symbol_table.concrete_classes:
        interface_name = golang_naming.interface_name(concrete_cls.name)

        cls_name_json = naming.json_model_type(concrete_cls.name)

        deserialization_function = golang_naming.function_name(
            Identifier(f"{concrete_cls.name}_from_jsonable")
        )

        deserialization_snippet = Stripped(
            f"""\
instance, err := aasjsonization.{deserialization_function}(
{I}jsonable,
)
if err != nil {{
{I}panic(
{II}fmt.Sprintf(
{III}"Failed to de-serialize an instance of {interface_name} " +
{III}"from %s: %s",
{III}pth, err.Error(),
{II}),
{I})
}}
var ok bool
result, ok = instance.(aastypes.{interface_name})
if !ok {{
{I}panic(
{II}fmt.Sprintf(
{III}"Expected to find an instance of {interface_name} at %s, " +
{III}"but got an instance of %T: %v",
{III}pth, instance, instance,
{II}),
{I})
}}"""
        )

        must_load_maximal_name = golang_naming.function_name(
            Identifier(f"must_load_maximal_{concrete_cls.name}")
        )

        must_load_minimal_name = golang_naming.function_name(
            Identifier(f"must_load_minimal_{concrete_cls.name}")
        )

        blocks.append(
            Stripped(
                f"""\
// Load a maximal example of [aastypes.{interface_name}] from
// the test data directory.
//
// If there is any error, panic.
func {must_load_maximal_name}(
) (result aastypes.{interface_name}) {{
{I}pth := path.Join(
{II}TestDataDir,
{II}"Json",
{II}"Expected",
{II}{golang_common.string_literal(cls_name_json)},
{II}"maximal.json",
{I})

{I}jsonable := MustReadJsonable(pth)

{I}{indent_but_first_line(deserialization_snippet, I)}
{I}return
}}"""
            )
        )

        blocks.append(
            Stripped(
                f"""\
// Load a minimal example of [aastypes.{interface_name}] from
// the test data directory.
//
// If there is any error, panic.
func {must_load_minimal_name}(
) (result aastypes.{interface_name}) {{
{I}pth := path.Join(
{II}TestDataDir,
{II}"Json",
{II}"Expected",
{II}{golang_common.string_literal(cls_name_json)},
{II}"minimal.json",
{I})

{I}jsonable := MustReadJsonable(pth)

{I}{indent_but_first_line(deserialization_snippet, I)}
{I}return
}}"""
            )
        )

    blocks.append(golang_common.WARNING)

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(block)

    writer.write("\n")

    return writer.getvalue()


assert generate.__doc__ is not None
assert generate.__doc__.strip().startswith(__doc__.strip())



"""Generate code for common functionality shared across the tests."""

import re
from typing import List

from aas_core_codegen.common import Stripped, Identifier, indent_but_first_line
from aas_core_codegen.java import (
    common as java_common,
)
from aas_core_codegen.java.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
    INDENT4 as IIII,
    INDENT5 as IIIII,
    INDENT6 as IIIIII,
)


def _package_to_upper_snake(
    package: java_common.PackageIdentifier,
) -> Identifier:
    """
    Convert a snake case dotted package into UPPER_SNAKE_CASE.

    >>> _package_to_upper_snake(java_common.PackageIdentifier("aas_core.aas3_1"))
    'AAS_CORE_AAS3_1'

    >>> _package_to_upper_snake(java_common.PackageIdentifier("simple_test"))
    'SIMPLE_TEST'
    """
    result = package.replace(".", "_")

    # Insert underscore between lowercase and uppercase (aasCore → aas_Core)
    result = re.sub(r"(?<=[a-z])(?=[A-Z])", "_", result)

    return Identifier(result.upper())


def generate(package: java_common.PackageIdentifier) -> List[java_common.JavaFile]:
    """Generate code for common functionality shared across the tests."""
    blocks = [
        Stripped(
            f"""\
public static final String RECORD_MODE_ENVIRONMENT_VARIABLE_NAME =
{I}"{_package_to_upper_snake(package)}_TESTS_RECORD_MODE";"""
        ),
        Stripped(
            f"""\
// NOTE (empwilli):
// It is tedious to record manually all the expected error messages.
// Therefore we include this variable to steer the automatic recording.
// We intentionally inter-twine the recording code with the test code
// to keep them close to each other so that they are easier to maintain.
public static final boolean RECORD_MODE =
{I}System.getenv("{_package_to_upper_snake(package)}_TESTS_RECORD_MODE") != null && (
{II}System.getenv(RECORD_MODE_ENVIRONMENT_VARIABLE_NAME).equalsIgnoreCase("true") ||
{II}System.getenv(RECORD_MODE_ENVIRONMENT_VARIABLE_NAME).equalsIgnoreCase("on") ||
{II}System.getenv(RECORD_MODE_ENVIRONMENT_VARIABLE_NAME).equalsIgnoreCase("1") ||
{II}System.getenv(RECORD_MODE_ENVIRONMENT_VARIABLE_NAME).equalsIgnoreCase("yes"));""",
        ),
        Stripped(
            f"""\
@SuppressWarnings("unchecked")
public static <T extends IClass> T mustFind(IClass container, Class<T> type) {{
{I}if (type.isInstance(container)) {{
{II}return type.cast(container);
{I}}}
{I}for (IClass current : container.descend()) {{
{II}if (type.isInstance(current)) {{
{III}return type.cast(current);
{II}}}
{I}}}
{I}throw new IllegalStateException("No instance of " + type.getSimpleName() + " could be found");
}}"""
        ),
        Stripped(
            f"""\
public static List<Path> findPaths(Path path, String fileExtension) throws IOException {{
{I}if (!Files.isDirectory(path)) {{
{II}throw new IllegalArgumentException("Path must be a directory!");
{I}}}

{I}List<Path> result;
{I}try (Stream<Path> walk = Files.walk(path)) {{
{II}result =
{III}walk.filter(p -> !Files.isDirectory(p))
{IIII}.filter(f -> f.toString().endsWith(fileExtension))
{IIII}.collect(Collectors.toList());
{I}}}
{I}return result;
}}""",
        ),
        Stripped(
            f"""\
public static List<Path> findDirs(Path path) throws IOException {{
{I}if (!Files.isDirectory(path)) {{
{II}throw new IllegalArgumentException("Path must be a directory!");
{I}}}

{I}List<Path> result;
{I}try (Stream<Path> walk = Files.walk(path)) {{
{II}result = walk.filter(p -> Files.isDirectory(p)).collect(Collectors.toList());
{I}}}
{I}return result;
}}""",
        ),
        Stripped(
            f"""\
public static <T> List<T> asList(Iterable<T> iterable) {{
{I}return StreamSupport.stream(iterable.spliterator(), false).collect(Collectors.toList());
}}"""
        ),
        Stripped(
            f"""\
public static void assertNoVerificationErrors(List<Reporting.Error> errors, Path path) {{
{I}if (!errors.isEmpty()) {{
{II}StringBuilder stringBuilder = new StringBuilder();
{II}stringBuilder
{III}.append("Expected no errors when verifying the instance de-serialized from ")
{III}.append(path.toString())
{III}.append(", ")
{III}.append("but got ")
{III}.append(errors.size())
{III}.append(" error(s):")
{III}.append(System.lineSeparator());
{II}for (Reporting.Error error : errors) {{
{III}stringBuilder
{IIII}.append(Reporting.generateJsonPath(error.getPathSegments()))
{IIII}.append(": ")
{IIII}.append(error.getCause());
{II}}}
{II}fail(stringBuilder.toString());
{I}}}
}}""",
        ),
        Stripped(
            f"""\
public static void assertEqualsExpectedOrRerecordVerificationErrors(
{I}List<Reporting.Error> errors, Path path) throws IOException {{
{I}if (errors.isEmpty()) {{
{II}fail("Expected at least one verification error when verifying " + path + ", but got none");
{I}}}
{I}final String got =
{II}errors.stream()
{III}.map(
{IIII}error ->
{IIIII}Reporting.generateJsonPath(error.getPathSegments()) + ": " + error.getCause())
{IIIIII}.collect(Collectors.joining(";\\n"));
{I}final Path errorsPath = Paths.get(path + ".errors");
{I}if (RECORD_MODE) {{
{II}Files.write(errorsPath, got.getBytes(StandardCharsets.UTF_8));
{I}}} else {{
{II}if (!Files.exists(errorsPath)) {{
{III}throw new FileNotFoundException(
{IIII}"The file with the recorded errors does not exist: "
{IIIII}+ errorsPath
{IIIII}+ "; maybe you want to set the environment variable "
{IIIII}+ Common.RECORD_MODE_ENVIRONMENT_VARIABLE_NAME);
{II}}}
{II}final String expected = String.join("\\n", Files.readAllLines(errorsPath));
{II}assertEquals(
{III}expected,
{III}got,
{III}"The expected verification errors do not match the actual ones for the file " + path);
{I}}}
}}"""
        ),
        Stripped(
            f"""\
public static String trace(IClass instance)
{{
{I}return instance.getClass().getName();
}}"""
        ),
    ]

    blocks_joined = "\n\n".join(blocks)

    content = f"""\
{java_common.WARNING}

package {package}.tests;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.fail;

import {package}.reporting.Reporting;
import {package}.types.model.IClass;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.List;
import java.util.stream.Collectors;
import java.util.stream.Stream;
import java.util.stream.StreamSupport;

/**
 * Provide common functionality to be re-used across different tests
 * such as reading of environment variables.
 */
public class Common {{
{I}public static String TEST_DATA_DIR = Paths.get("test_data").toAbsolutePath().toString();

{I}{indent_but_first_line(blocks_joined, I)}
}} // class Common

{java_common.WARNING}
"""

    return [java_common.JavaFile("Common.java", content)]


assert generate.__doc__ is not None
assert generate.__doc__.strip().startswith(__doc__.strip())



"""Provide data structures for the constraint inferences."""
import sys
from typing import Mapping, Sequence, Optional, Final, MutableMapping

if sys.version_info >= (3, 10):
    from typing import TypeAlias
else:
    from typing_extensions import TypeAlias

from icontract import require  # pylint: disable=wrong-import-position

from aas_core_codegen import intermediate  # pylint: disable=wrong-import-position


class LenConstraint:
    """
    Represent the inferred constraint on the ``len`` of something.

    Both bounds are inclusive: ``min_value ≤ len ≤ max_value``.
    """

    min_value: Final[Optional[int]]
    max_value: Final[Optional[int]]

    # fmt: off
    @require(
        lambda min_value, max_value:
        not (min_value is not None and max_value is not None)
        or 0 < min_value <= max_value
    )
    # fmt: on
    def __init__(self, min_value: Optional[int], max_value: Optional[int]) -> None:
        """Initialize with the given values."""
        self.min_value = min_value
        self.max_value = max_value

    def equals(self, other: Optional["LenConstraint"]) -> bool:
        """Return true if the other constraint equals semantically ours."""
        if other is None:
            return False

        return self.min_value == other.min_value and self.max_value == other.max_value

    def __str__(self) -> str:
        return (
            f"{self.__class__.__name__}("
            f"min_value={self.min_value!r}, max_value={self.max_value!r})"
        )


class PatternConstraint:
    """Constrain a string to comply to a regular expression."""

    pattern: Final[str]

    def __init__(self, pattern: str) -> None:
        """Initialize with the given values."""
        self.pattern = pattern

    def equals(self, other: Optional["PatternConstraint"]) -> bool:
        """Return true if the other constraint equals semantically ours."""
        if other is None:
            return False

        return self.pattern == other.pattern

    def __repr__(self) -> str:
        """Represent the constraint with the pattern."""
        return (
            f"<{self.__class__.__name__} at 0x{id(self):x} "
            f"with pattern={self.pattern!r}>"
        )


class SetOfPrimitivesConstraint:
    """Constrain a primitive value to be a member of a pre-defined set of values."""

    a_type: Final[intermediate.PrimitiveType]
    literals: Final[Sequence[intermediate.PrimitiveSetLiteral]]

    # fmt: off
    @require(
        lambda a_type, literals:
        all(
            literal.a_type is a_type
            for literal in literals
        )
    )
    # fmt: on
    def __init__(
        self,
        a_type: intermediate.PrimitiveType,
        literals: Sequence[intermediate.PrimitiveSetLiteral],
    ) -> None:
        """Initialize with the given values."""
        self.a_type = a_type
        self.literals = literals

    def equals(self, other: Optional["SetOfPrimitivesConstraint"]) -> bool:
        """Return true if the other constraint equals semantically ours."""
        if other is None:
            return False

        return (
            self.a_type == other.a_type
            and len(self.literals) == len(other.literals)
            and all(
                id(that_literal) == id(other_literal)
                for that_literal, other_literal in zip(self.literals, other.literals)
            )
        )

    def __repr__(self) -> str:
        """Represent the constraint with the pattern."""
        return f"<{self.__class__.__name__} at 0x{id(self):x}>"


class SetOfEnumerationLiteralsConstraint:
    """Constrain a value to be a member of a pre-defined set of values."""

    enumeration: Final[intermediate.Enumeration]
    literals: Final[Sequence[intermediate.EnumerationLiteral]]

    # fmt: off
    @require(
        lambda enumeration, literals:
        all(
            id(literal) in enumeration.literal_id_set
            for literal in literals
        )
    )
    # fmt: on
    def __init__(
        self,
        enumeration: intermediate.Enumeration,
        literals: Sequence[intermediate.EnumerationLiteral],
    ) -> None:
        """Initialize with the given values."""
        self.enumeration = enumeration
        self.literals = literals

    @require(
        lambda self, other: not (other is not None)
        or (other.enumeration is self.enumeration),
        "Both self and other must refer to the same enumeration for "
        "an equality comparison to make sense.",
    )
    def equals(self, other: Optional["SetOfEnumerationLiteralsConstraint"]) -> bool:
        """Return true if the other constraint equals semantically ours."""
        if other is None:
            return False

        return len(self.literals) == len(other.literals) and all(
            id(that_literal) == id(other_literal)
            for that_literal, other_literal in zip(self.literals, other.literals)
        )

    def __repr__(self) -> str:
        """Represent the constraint with the pattern."""
        return f"<{self.__class__.__name__} at 0x{id(self):x}>"


class Constraints:
    """
    Represent all the inferred constraints for a value.

    The constraints coming from the constrained primitives are in-lined and hence also
    included in this representation.
    """

    len_constraint: Final[Optional[LenConstraint]]
    patterns: Final[Optional[Sequence[PatternConstraint]]]
    set_of_primitives: Final[Optional[SetOfPrimitivesConstraint]]
    set_of_enumeration_literals: Final[Optional[SetOfEnumerationLiteralsConstraint]]

    @require(lambda patterns: not (patterns is not None) or len(patterns) > 0)
    def __init__(
        self,
        len_constraint: Optional[LenConstraint] = None,
        patterns: Optional[Sequence[PatternConstraint]] = None,
        set_of_primitives: Optional[SetOfPrimitivesConstraint] = None,
        set_of_enumeration_literals: Optional[
            SetOfEnumerationLiteralsConstraint
        ] = None,
    ) -> None:
        """Initialize with the given values."""
        self.len_constraint = len_constraint
        self.patterns = patterns
        self.set_of_primitives = set_of_primitives
        # fmt: off
        self.set_of_enumeration_literals = (
            set_of_enumeration_literals
        )
        # fmt: on

    def is_empty(self) -> bool:
        """Return True if no constraints are set."""
        return (
            self.len_constraint is None
            and self.patterns is None
            and self.set_of_primitives is None
            and self.set_of_enumeration_literals is None
        )


#: Represent the constraints inferred for the given value in a class.
ConstraintsByValue: TypeAlias = Mapping[intermediate.TypeAnnotationUnion, Constraints]

MutableConstraintsByValue: TypeAlias = MutableMapping[
    intermediate.TypeAnnotationUnion, Constraints
]



"""Generate code to test the verification of pos. and neg. cases."""

import io
from typing import List

from icontract import ensure

from aas_core_codegen import intermediate, naming
from aas_core_codegen.common import (
    Stripped,
)
from aas_core_codegen.cpp import common as cpp_common, naming as cpp_naming
from aas_core_codegen.cpp.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
    INDENT4 as IIII,
    INDENT5 as IIIII,
    INDENT6 as IIIIII,
)


# fmt: off
@ensure(
    lambda result:
    result.endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate_implementation(
    symbol_table: intermediate.SymbolTable, library_namespace: Stripped
) -> str:
    """Generate implementation to test the verification of pos. and neg. cases."""
    include_prefix_path = cpp_common.generate_include_prefix_path(library_namespace)

    blocks = [
        cpp_common.WARNING,
        Stripped(
            f"""\
#include "./common.hpp"
#include "./common_xmlization.hpp"

#include <{include_prefix_path}/verification.hpp>
#include <{include_prefix_path}/xmlization.hpp>

#define CATCH_CONFIG_MAIN
#include <catch2/catch.hpp>

namespace aas = {library_namespace};"""
        ),
        Stripped(
            f"""\
void AssertNoVerificationError(
{I}const std::filesystem::path& xml_path
) {{
{I}std::shared_ptr<
{II}aas::types::IClass
{I}> instance(
{II}test::common::xmlization::MustDeserializeFile(xml_path)
{I});

{I}std::vector<std::string> error_messages;
{I}for (
{II}const aas::verification::Error &error
{II}: aas::verification::RecursiveVerification(instance)
{I}) {{
{II}error_messages.emplace_back(
{III}aas::common::Concat(
{IIII}aas::common::WstringToUtf8(error.path.ToWstring()),
{IIII}": ",
{IIII}aas::common::WstringToUtf8(error.cause)
{III})
{II});
{I}}}

{I}if (!error_messages.empty()) {{
{II}std::vector<std::string> parts;
{II}parts.emplace_back("Expected no error messages from ");
{II}parts.emplace_back(xml_path.string());
{II}parts.emplace_back(", but got:\\n");
{II}parts.emplace_back(test::common::JoinStrings(error_messages, "\\n"));

{II}INFO(test::common::JoinStrings(parts, ""))
{II}CHECK(error_messages.empty());
{I}}}
}}"""
        ),
        Stripped(
            f"""\
const std::filesystem::path& DetermineXmlDir() {{
{I}static aas::common::optional<std::filesystem::path> result;
{I}if (!result.has_value()) {{
{II}result = test::common::DetermineTestDataDir() / "Xml";
{I}}}

{I}return *result;
}}"""
        ),
        Stripped(
            f"""\
const std::filesystem::path& DetermineErrorDir() {{
{I}static aas::common::optional<std::filesystem::path> result;
{I}if (!result.has_value()) {{
{II}result = test::common::DetermineTestDataDir() / "VerificationError";
{I}}}

{I}return *result;
}}"""
        ),
        Stripped(
            f"""\
void AssertVerificationFailure(
{I}const std::filesystem::path& path,
{I}const std::filesystem::path& error_path
) {{
{I}std::ifstream ifs(path, std::ios::binary);

{I}aas::common::expected<
{II}std::shared_ptr<aas::types::IClass>,
{II}aas::xmlization::DeserializationError
{I}> deserialized = aas::xmlization::From(
{II}ifs
{I});

{I}if (!deserialized.has_value()) {{
{II}INFO(
{III}aas::common::Concat(
{IIII}"Expected to de-serialize ",
{IIII}path.string(),
{IIII}", but the de-serialization failed: ",
{IIII}aas::common::WstringToUtf8(deserialized.error().path.ToWstring()),
{IIII}aas::common::WstringToUtf8(deserialized.error().cause)
{III})
{II})
{II}REQUIRE(!deserialized.has_value());
{I}}}

{I}std::vector<std::string> error_messages;
{I}for (
{II}const aas::verification::Error& error
{II}: aas::verification::RecursiveVerification(deserialized.value())
{II}) {{
{II}error_messages.emplace_back(
{III}aas::common::Concat(
{IIII}aas::common::WstringToUtf8(error.path.ToWstring()),
{IIII}": ",
{IIII}aas::common::WstringToUtf8(error.cause)
{III})
{II});
{I}}}

{I}if (error_messages.empty()) {{
{II}INFO(
{III}aas::common::Concat(
{IIII}"Expected error messages from ",
{IIII}path.string(),
{IIII}", but got none"
{III})
{II})
{II}REQUIRE(!error_messages.empty());
{I}}}

{I}const std::string joined_error_messages = test::common::JoinStrings(
{II}error_messages,
{II}"\\n"
{I});

{I}test::common::AssertContentEqualsExpectedOrRecord(
{II}joined_error_messages,
{II}error_path
{I});
}}"""
        ),
    ]  # type: List[Stripped]

    for concrete_cls in symbol_table.concrete_classes:
        xml_class_name = naming.xml_class_name(concrete_cls.name)

        cls_name = cpp_naming.class_name(concrete_cls.name)

        blocks.append(
            Stripped(
                f"""\
TEST_CASE("Test verification of a valid {cls_name}") {{
{I}const std::deque<std::filesystem::path> paths(
{II}test::common::FindFilesBySuffixRecursively(
{III}DetermineXmlDir()
{IIII}/ "Expected"
{IIII}/ {cpp_common.string_literal(xml_class_name)},
{III}".xml"
{II})
{I});

{I}for (const std::filesystem::path& path : paths) {{
{II}AssertNoVerificationError(path);
{I}}}
}}"""
            )
        )

        blocks.append(
            Stripped(
                f"""\
TEST_CASE("Test verification of invalid cases for {cls_name}") {{
{I}for (
{II}const std::filesystem::path& causeDir
{II}: test::common::ListSubdirectories(
{III}DetermineXmlDir()
{IIII}/ "Unexpected"
{IIII}/ "Invalid"
{II})
{I}) {{
{II}for (
{III}const std::filesystem::path& path
{III}: test::common::FindFilesBySuffixRecursively(
{IIII}causeDir / {cpp_common.string_literal(xml_class_name)},
{IIII}".xml"
{III})
{II}) {{
{III}const std::filesystem::path parent(
{IIII}(
{IIIII}DetermineErrorDir()
{IIIIII}/ std::filesystem::relative(path, DetermineXmlDir())
{IIII}).parent_path()
{III});

{III}const std::filesystem::path error_path(
{IIII}parent
{IIIII}/ (path.filename().string() + ".errors")
{III});

{III}AssertVerificationFailure(path, error_path);
{II}}}
{I}}}
}}"""
            )
        )

    blocks.append(cpp_common.WARNING)

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(block)

    writer.write("\n")

    return writer.getvalue()


assert generate_implementation.__doc__ is not None
cpp_common.assert_module_docstring_and_generate_implementation_consistent(
    module_doc=__doc__, generate_implementation_doc=generate_implementation.__doc__
)



"""Generate code to test the base64 encoding and decoding."""

import io

from icontract import ensure

from aas_core_codegen.common import (
    Stripped,
)
from aas_core_codegen.cpp import common as cpp_common
from aas_core_codegen.cpp.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
    INDENT4 as IIII,
)


# fmt: off
@ensure(
    lambda result:
    result.endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate_implementation(library_namespace: Stripped) -> str:
    """Generate implementation to test the base64 encoding and decoding."""
    include_prefix_path = cpp_common.generate_include_prefix_path(library_namespace)

    blocks = [
        cpp_common.WARNING,
        Stripped(
            """\
/**
 * Test base64 encoding and decoding.
 */"""
        ),
        Stripped(
            f'''\
#include "{include_prefix_path}/stringification.hpp"'''
        ),
        Stripped(
            """\
#define CATCH_CONFIG_MAIN
#include <catch2/catch.hpp>"""
        ),
        Stripped(
            f"""\
namespace aas = {library_namespace};"""
        ),
        Stripped(
            f"""\
std::string BytesToString(const std::vector<uint8_t>& bytes) {{
{I}std::string result;
{I}result.resize(bytes.size());
{I}for (size_t i = 0; i < bytes.size(); ++i
{II}) {{
{II}result[i] = bytes[i];
{I}}}
{I}return result;
}}"""
        ),
        Stripped(
            f"""\
std::vector<uint8_t> StringToBytes(const std::string& text) {{
{I}std::vector<uint8_t> result;
{I}result.resize(text.size());
{I}for (size_t i = 0; i < text.size(); ++i
{II}) {{
{II}result[i] = text[i];
{I}}}
{I}return result;
}}"""
        ),
        Stripped(
            f"""\
void AssertEncodeDecode(
{I}const std::string& text,
{I}const std::string& expected_encoded
) {{
{I}const std::string encoded = aas::stringification::Base64Encode(
{II}StringToBytes(text)
{I});

{I}REQUIRE(
{II}expected_encoded
{III}== encoded
{I});

{I}aas::common::expected<
{II}std::vector<std::uint8_t>,
{II}std::string
{I}> bytes = aas::stringification::Base64Decode(encoded);

{I}REQUIRE(bytes.has_value());

{I}REQUIRE(
{II}BytesToString(bytes.value())
{III}== text
{I});
}}"""
        ),
        Stripped(
            f"""\
TEST_CASE("Test '' is encoded ''") {{
{I}AssertEncodeDecode(
{II}"",
{II}""
{I});
}}"""
        ),
        Stripped(
            f"""\
TEST_CASE("Test 'f' is encoded 'Zg=='") {{
{I}AssertEncodeDecode(
{II}"f",
{II}"Zg=="
{I});
}}"""
        ),
        Stripped(
            f"""\
TEST_CASE("Test 'fo' is encoded 'Zm8='") {{
{I}AssertEncodeDecode(
{II}"fo",
{II}"Zm8="
{I});
}}"""
        ),
        Stripped(
            f"""\
TEST_CASE("Test 'foo' is encoded 'Zm9v'") {{
{I}AssertEncodeDecode(
{II}"foo",
{II}"Zm9v"
{I});
}}"""
        ),
        Stripped(
            f"""\
TEST_CASE("Test 'foob' is encoded 'Zm9vYg=='") {{
{I}AssertEncodeDecode(
{II}"foob",
{II}"Zm9vYg=="
{I});
}}"""
        ),
        Stripped(
            f"""\
TEST_CASE("Test 'fooba' is encoded 'Zm9vYmE='") {{
{I}AssertEncodeDecode(
{II}"fooba",
{II}"Zm9vYmE="
{I});
}}"""
        ),
        Stripped(
            f"""\
TEST_CASE("Test 'foobar' is encoded 'Zm9vYmFy'") {{
{I}AssertEncodeDecode(
{II}"foobar",
{II}"Zm9vYmFy"
{I});
}}"""
        ),
        Stripped(
            f"""\
TEST_CASE("Test unexpected padding in the middle") {{
{I}const std::string encoded = "Zm9vYmFy";
{I}for (size_t i = 0; i < encoded.size() - 1; ++i
{II}) {{
{II}const std::string bad_encoded(
{III}encoded.substr(0, i)
{IIII}+ "="
{IIII}+ encoded.substr(i + 1, encoded.size() - i)
{II});

{II}aas::common::expected<
{III}std::vector<std::uint8_t>,
{III}std::string
{II}> bytes = aas::stringification::Base64Decode(bad_encoded);

{II}REQUIRE(!bytes.has_value());

{II}std::stringstream ss;
{II}ss
{III}<< "Expected a valid character from base64-encoded string, "
{III}<< "but got at index " << i << ": 61 (code: 61)";

{II}REQUIRE(
{III}bytes.error()
{IIII}== ss.str()
{II});
{I}}}
}}"""
        ),
        Stripped(
            f"""\
// NOTE (mristin):
// The following checks come from: https://eprint.iacr.org/2022/361.pdf
TEST_CASE("Test 'Hello' is encoded as 'SGVsbG8='") {{
{I}AssertEncodeDecode(
{II}"Hello",
{II}"SGVsbG8="
{I});
}}"""
        ),
        Stripped(
            f"""\
// NOTE (mristin):
// This is not a test case, but we merely document that there is a possible
// attack vector where different strings encode the *same* byte sequence.
TEST_CASE("Test that our implementation suffers from padding inconsistency of 'Hello' as 'SGVsbG9='") {{
{I}const std::string encoded = "SGVsbG9=";

{I}aas::common::expected<
{II}std::vector<std::uint8_t>,
{II}std::string
{I}> bytes = aas::stringification::Base64Decode(encoded);

{I}REQUIRE(bytes.has_value());

{I}REQUIRE(
{II}BytesToString(bytes.value())
{III}== "Hello"
{I});
}}"""
        ),
        Stripped(
            f"""\
void AssertDecode(
{I}const std::string& encoded,
{I}const std::vector<std::uint8_t>& expected_decoded
) {{
{I}aas::common::expected<
{II}std::vector<std::uint8_t>,
{II}std::string
{I}> bytes = aas::stringification::Base64Decode(encoded);

{I}INFO(encoded)
{I}REQUIRE(bytes.has_value());

{I}INFO(encoded)
{I}REQUIRE(bytes->size() == expected_decoded.size());

{I}for (size_t i = 0; i < bytes->size(); ++i
{II}) {{
{II}INFO(encoded + " is not correctly decoded at " + std::to_string(i))
{II}REQUIRE(bytes->at(i) == expected_decoded[i]);
{I}}}
}}"""
        ),
        Stripped(
            f"""\
// NOTE (mristin):
// The following tests have been scraped from:
// https://github.com/rickkas7/Base64RK/blob/183d20e62b96ef9b0230c80c7a172715ca6f661e/test/unit-test/unit-test.cpp
TEST_CASE("Test table from rickkas7/Base64RK") {{
{I}AssertDecode(
{II}"",
{II}{{}}
{I});

{I}AssertDecode(
{II}"HA==",
{II}{{28}}
{I});

{I}AssertDecode(
{II}"rtA=",
{II}{{174, 208}}
{I});

{I}AssertDecode(
{II}"MuDQ",
{II}{{50, 224, 208}}
{I});

{I}AssertDecode(
{II}"LRLrBg==",
{II}{{45, 18, 235, 6}}
{I});

{I}AssertDecode(
{II}"dceZ3aA=",
{II}{{117, 199, 153, 221, 160}}
{I});

{I}AssertDecode(
{II}"jQy8MK0t",
{II}{{141, 12, 188, 48, 173, 45}}
{I});

{I}AssertDecode(
{II}"RY+UfdUDlA==",
{II}{{69, 143, 148, 125, 213, 3, 148}}
{I});

{I}AssertDecode(
{II}"+6KmPIGDLg0=",
{II}{{251, 162, 166, 60, 129, 131, 46, 13}}
{I});

{I}AssertDecode(
{II}"zyjM/k1HPQiA",
{II}{{207, 40, 204, 254, 77, 71, 61, 8, 128}}
{I});

{I}AssertDecode(
{II}"nJbVrnU7RtzSPw==",
{II}{{156, 150, 213, 174, 117, 59, 70, 220, 210, 63}}
{I});

{I}AssertDecode(
{II}"wDatE35UGvqnBTI=",
{II}{{192, 54, 173, 19, 126, 84, 26, 250, 167, 5, 50}}
{I});

{I}AssertDecode(
{II}"FAEPTeCNmS+6qsjj",
{II}{{20, 1, 15, 77, 224, 141, 153, 47, 186, 170, 200, 227}}
{I});

{I}AssertDecode(
{II}"VaaCqcF1wlHuQLTkfw==",
{II}{{85, 166, 130, 169, 193, 117, 194, 81, 238, 64, 180, 228, 127}}
{I});

{I}AssertDecode(
{II}"QYKREfdjniR1CMABweo=",
{II}{{65, 130, 145, 17, 247, 99, 158, 36, 117, 8, 192, 1, 193, 234}}
{I});

{I}AssertDecode(
{II}"CTf9s8td7Tr8yKiRczUf",
{II}{{9, 55, 253, 179, 203, 93, 237, 58, 252, 200, 168, 145, 115, 53, 31}}
{I});

{I}AssertDecode(
{II}"/kyMTRA0d+W64W21hiLlmesBEH4RDg0MivjmPEmAdMzP",
{II}{{254, 76, 140, 77, 16, 52, 119, 229, 186, 225, 109, 181, 134, 34,
{II} 229, 153, 235, 1, 16, 126, 17, 14, 13, 12, 138,
{II} 248, 230, 60, 73, 128, 116, 204, 207}}
{I});

{I}AssertDecode(
{II}"m5/QqN3YcQg6iE1dVjJQWw5mCjYCnsNLKPVUsXQ269l86M1xbyHn+/aIKafJ6hRT4UOD0mH"
{II}"CAfye/VSHdwH+C7bQ",
{II}{{155, 159, 208, 168, 221, 216, 113, 8, 58, 136, 77, 93, 86, 50, 80,
{II} 91, 14, 102, 10, 54, 2, 158, 195, 75, 40, 245,
{II} 84, 177, 116, 54, 235, 217, 124, 232, 205, 113, 111, 33, 231, 251, 246, 136, 41,
{II} 167, 201, 234, 20, 83, 225, 67,
{II} 131, 210, 97, 194, 1, 252, 158, 253, 84, 135, 119, 1, 254, 11, 182, 208}}
{I});

{I}AssertDecode(
{II}"ZSsh3WyA3d00CyimeMZPAI1/t+LF6DQdiAEQG9hHlfuDXhZ2Gf/Wd1u2y+RDFli0oSq2fbfv"
{II}"utgXCxGQfMmxA1NEg/k0vQIcSqzWdX7571oWzKHz6KZnlhTwPF27QznnAWbdwkyiCFlIlpL3tYbyzbMT"
{II}"m0ReUtY4AW9tTpvYSQdQ8w==",
{II}{{101, 43, 33, 221, 108, 128, 221, 221, 52, 11, 40, 166, 120, 198,
{II} 79, 0, 141, 127, 183, 226, 197, 232, 52, 29, 136,
{II} 1, 16, 27, 216, 71, 149, 251, 131, 94, 22, 118, 25, 255, 214, 119, 91, 182, 203,
{II} 228, 67, 22, 88, 180, 161, 42,
{II} 182, 125, 183, 239, 186, 216, 23, 11, 17, 144, 124, 201, 177, 3, 83, 68, 131,
{II} 249, 52, 189, 2, 28, 74, 172, 214,
{II} 117, 126, 249, 239, 90, 22, 204, 161, 243, 232, 166, 103, 150, 20, 240, 60, 93,
{II} 187, 67, 57, 231, 1, 102, 221, 194,
{II} 76, 162, 8, 89, 72, 150, 146, 247, 181, 134, 242, 205, 179, 19, 155, 68, 94, 82,
{II} 214, 56, 1, 111, 109, 78, 155,
{II} 216, 73, 7, 80, 243}}
{I});

{I}AssertDecode(
{II}"seCV0jbCe42d4GMBvXIl5vqVoZ6K4gUSg6nKgi5tZiNxx22yYon5ReOC6PZpZWHqeJcfKVo"
{II}"UD/Za1ieLVjROriNfP+Lxf6Tbroz/EXh2YYOFI+d110oObkYqbgcl6ovS4CRH40Sb7L89Uwu9WDzQA"
{II}"215IEepOn1wdBCZyEryWY2PGdNcs+Pft9sSNtBQ/QtMRZYEMO5JpRmFyFDPVLZA+5xM/3Dj65odv1b"
{II}"ZdtszJq0Gbw43ww0t7EK0J21uBy89Z2R1N1kjQGpqKu4lRhQIgCKDOHhnfZW2gtkmZhiqby9xT44Ep1"
{II}"AMrmZKgW3MDPzIX2Ez16mUv3Gb5JHrI/w1sQ==",
{II}{{177, 224, 149, 210, 54, 194, 123, 141, 157, 224, 99, 1, 189,
{II} 114, 37, 230, 250, 149, 161, 158, 138, 226, 5, 18,
{II} 131, 169, 202, 130, 46, 109, 102, 35, 113, 199, 109, 178, 98, 137, 249, 69, 227,
{II} 130, 232, 246, 105, 101, 97, 234,
{II} 120, 151, 31, 41, 90, 20, 15, 246, 90, 214, 39, 139, 86, 52, 78, 174, 35, 95, 63,
{II} 226, 241, 127, 164, 219, 174,
{II} 140, 255, 17, 120, 118, 97, 131, 133, 35, 231, 117, 215, 74, 14, 110, 70, 42,
{II} 110, 7, 37, 234, 139, 210, 224, 36,
{II} 71, 227, 68, 155, 236, 191, 61, 83, 11, 189, 88, 60, 208, 3, 109, 121, 32, 71,
{II} 169, 58, 125, 112, 116, 16, 153,
{II} 200, 74, 242, 89, 141, 143, 25, 211, 92, 179, 227, 223, 183, 219, 18, 54, 208,
{II} 80, 253, 11, 76, 69, 150, 4, 48,
{II} 238, 73, 165, 25, 133, 200, 80, 207, 84, 182, 64, 251, 156, 76, 255, 112, 227,
{II} 235, 154, 29, 191, 86, 217, 118,
{II} 219, 51, 38, 173, 6, 111, 14, 55, 195, 13, 45, 236, 66, 180, 39, 109, 110, 7,
{II} 47, 61, 103, 100, 117, 55, 89, 35,
{II} 64, 106, 106, 42, 238, 37, 70, 20, 8, 128, 34, 131, 56, 120, 103, 125, 149,
{II} 182, 130, 217, 38, 102, 24, 170, 111,
{II} 47, 113, 79, 142, 4, 167, 80, 12, 174, 102, 74, 129, 109, 204, 12, 252, 200,
{II} 95, 97, 51, 215, 169, 148, 191, 113,
{II} 155, 228, 145, 235, 35, 252, 53, 177}}
{I});
}}"""
        ),
        cpp_common.WARNING,
    ]

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(block)

    writer.write("\n")

    return writer.getvalue()


assert generate_implementation.__doc__ is not None
cpp_common.assert_module_docstring_and_generate_implementation_consistent(
    module_doc=__doc__, generate_implementation_doc=generate_implementation.__doc__
)



"""Generate the functions used across descent tests."""

import io
from typing import List

from icontract import ensure

from aas_core_codegen.common import (
    Stripped,
)
from aas_core_codegen.golang import common as golang_common
from aas_core_codegen.golang.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
    INDENT4 as IIII,
    INDENT5 as IIIII,
)


# fmt: off
@ensure(
    lambda result: result.endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate(repo_url: Stripped) -> str:
    """Generate the functions used across descent tests."""
    blocks = [
        Stripped(
            """\
package types_descend_test"""
        ),
        golang_common.WARNING,
        Stripped(
            f"""\
import (
{I}"fmt"
{I}"os"
{I}"path/filepath"
{I}"strings"
{I}aastesting "{repo_url}/aastesting"
{I}aastypes "{repo_url}/types"
)"""
        ),
        Stripped(
            f"""\
// Trace the `instance` and compare the trace against the golden one from
// the test data, or re-record the trace if [aastesting.RecordMode] is set.
//
// If `onlyOnce`, trace the `instance` with [aastypes.DescendOnce]. Otherwise,
// trace with [aastypes.Descend].
//
// If we are comparing, and not recording, return the error message if
// the expected and the obtained trace differ.
func compareOrRerecordTrace(
{I}instance aastypes.IClass,
{I}expectedPath string,
{I}onlyOnce bool,
) (message *string) {{
{I}lines := []string{{aastesting.TraceMark(instance)}}

{I}if onlyOnce {{
{II}instance.DescendOnce(func(descendant aastypes.IClass) (abort bool) {{
{III}lines = append(lines, aastesting.TraceMark(descendant))
{III}return
{II}}})
{I}}} else {{
{II}instance.Descend(func(descendant aastypes.IClass) (abort bool) {{
{III}lines = append(lines, aastesting.TraceMark(descendant))
{III}return
{II}}})
{I}}}

{I}got := strings.Join(lines, "\\n")

{I}// Add a new line for POSIX systems.
{I}got += "\\n"

{I}if aastesting.RecordMode {{
{II}parent := filepath.Dir(expectedPath)
{II}err := os.MkdirAll(parent, os.ModePerm)
{II}if err != nil {{
{III}panic(
{IIII}fmt.Sprintf(
{IIIII}"Failed to create the directory %s: %s", parent, err.Error(),
{IIII}),
{III})
{II}}}

{II}err = os.WriteFile(expectedPath, []byte(got), 0644)
{II}if err != nil {{
{III}panic(
{IIII}fmt.Sprintf(
{IIIII}"Failed to write to the file %s: %s", expectedPath, err.Error(),
{IIII}),
{III})
{II}}}
{I}}} else {{
{II}bb, err := os.ReadFile(expectedPath)
{II}if err != nil {{
{III}panic(
{IIII}fmt.Sprintf(
{IIIII}"Failed to read from file %s: %s", expectedPath, err.Error(),
{IIII}),
{III})
{II}}}

{II}expected := string(bb)

{II}// NOTE (mristin):
{II}// Git automatically strips and adds `\\r`, so we have to remove it here
{II}// to obtain a canonical text.
{II}expected = strings.Replace(expected, "\\r", "", -1)

{II}if expected != got {{
{III}text := fmt.Sprintf(
{IIII}"What we got differs from the expected in %s. "+
{IIIII}"We got:\\n%s\\nWe expected:\\n%s",
{IIII}expectedPath, got, expected,
{III})
{III}message = &text
{II}}}
{I}}}
{I}return
}}"""
        ),
        golang_common.WARNING,
    ]  # type: List[Stripped]

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(block)

    writer.write("\n")

    return writer.getvalue()


assert generate.__doc__ is not None
assert generate.__doc__.strip().startswith(__doc__.strip())



"""Generate C++ code based on the meta-model."""

import pathlib
from typing import TextIO, Sequence, Tuple, Callable, Optional, List

from aas_core_codegen import run, intermediate, specific_implementations, cpp
from aas_core_codegen.common import Stripped, Error
from aas_core_codegen.cpp import lib as cpp_lib, tests as cpp_tests


def execute(context: run.Context, stdout: TextIO, stderr: TextIO) -> int:
    """Generate code."""
    verified_ir_table, errors = cpp_lib.verify_for_types(
        symbol_table=context.symbol_table
    )

    assert verified_ir_table is not None

    unsupported_contracts_errors = (
        intermediate.errors_if_contracts_for_functions_or_methods_defined(
            verified_ir_table
        )
    )
    if unsupported_contracts_errors is not None:
        run.write_error_report(
            message=f"We do not support pre and post-conditions and snapshots "
            f"at the moment. Please notify the developers if you need this "
            f"feature (based on meta-model {context.model_path})",
            errors=[
                context.lineno_columner.error_message(error)
                for error in unsupported_contracts_errors
            ],
            stderr=stderr,
        )
        return 1

    unsupported_methods_errors = (
        intermediate.errors_if_non_implementation_specific_methods(verified_ir_table)
    )
    if unsupported_methods_errors is not None:
        run.write_error_report(
            message=f"We added some support for understood methods already and keep "
            f"maintaining it as it is only a matter of time when we will "
            f"introduce their transpilation. Introducing them after the fact "
            f"would have been much more difficult.\n"
            f"\n"
            f"At the given moment, however, we deliberately focus only on "
            f"implementation-specific methods. "
            f"(based on meta-model {context.model_path})",
            errors=[
                context.lineno_columner.error_message(error)
                for error in unsupported_methods_errors
            ],
            stderr=stderr,
        )
        return 1

    # region Namespace

    namespace_key = specific_implementations.ImplementationKey("namespace.txt")
    namespace_text = context.spec_impls.get(namespace_key, None)
    if namespace_text is None:
        stderr.write(f"The namespace snippet is missing: {namespace_key}\n")
        return 1

    library_namespace = Stripped(namespace_text.strip())

    # endregion

    try:
        context.output_dir.mkdir(exist_ok=True, parents=True)
    except Exception as exception:
        stderr.write(
            f"Failed to create the output directory {context.output_dir}: {exception}"
        )
        return 1

    src_dir = pathlib.Path("src")
    assert not src_dir.is_absolute()

    include_dir = pathlib.Path("include") / library_namespace.replace("::", "/")
    assert not include_dir.is_absolute()

    test_dir = pathlib.Path("test")
    assert not test_dir.is_absolute()

    rel_paths_generators: Sequence[
        Tuple[pathlib.Path, Callable[[], Tuple[Optional[str], Optional[List[Error]]]]]
    ] = [
        (
            include_dir / "common.hpp",
            lambda: (
                cpp_lib.generate_common_header(library_namespace=library_namespace),
                None,
            ),
        ),
        (
            src_dir / "common.cpp",
            lambda: (
                cpp_lib.generate_common_implementation(
                    library_namespace=library_namespace
                ),
                None,
            ),
        ),
        (
            include_dir / "constants.hpp",
            lambda: cpp_lib.generate_constants_header(
                symbol_table=context.symbol_table, library_namespace=library_namespace
            ),
        ),
        (
            src_dir / "constants.cpp",
            lambda: cpp_lib.generate_constants_implementation(
                symbol_table=context.symbol_table, library_namespace=library_namespace
            ),
        ),
        (
            include_dir / "enhancing.hpp",
            lambda: cpp_lib.generate_enhancing_header(
                symbol_table=context.symbol_table, library_namespace=library_namespace
            ),
        ),
        (
            include_dir / "iteration.hpp",
            lambda: cpp_lib.generate_iteration_header(
                symbol_table=context.symbol_table, library_namespace=library_namespace
            ),
        ),
        (
            src_dir / "iteration.cpp",
            lambda: cpp_lib.generate_iteration_implementation(
                symbol_table=context.symbol_table, library_namespace=library_namespace
            ),
        ),
        (
            include_dir / "jsonization.hpp",
            lambda: (
                cpp_lib.generate_jsonization_header(
                    symbol_table=verified_ir_table, library_namespace=library_namespace
                ),
                None,
            ),
        ),
        (
            src_dir / "jsonization.cpp",
            lambda: cpp_lib.generate_jsonization_implementation(
                symbol_table=context.symbol_table,
                spec_impls=context.spec_impls,
                library_namespace=library_namespace,
            ),
        ),
        (
            include_dir / "pattern.hpp",
            lambda: (
                cpp_lib.generate_pattern_header(
                    symbol_table=context.symbol_table,
                    library_namespace=library_namespace,
                ),
                None,
            ),
        ),
        (
            src_dir / "pattern.cpp",
            lambda: cpp_lib.generate_pattern_implementation(
                symbol_table=context.symbol_table,
                library_namespace=library_namespace,
            ),
        ),
        (
            include_dir / "revm.hpp",
            lambda: (
                cpp_lib.generate_revm_header(library_namespace=library_namespace),
                None,
            ),
        ),
        (
            src_dir / "revm.cpp",
            lambda: (
                cpp_lib.generate_revm_implementation(
                    library_namespace=library_namespace,
                ),
                None,
            ),
        ),
        (
            include_dir / "stringification.hpp",
            lambda: (
                cpp_lib.generate_stringification_header(
                    symbol_table=context.symbol_table,
                    library_namespace=library_namespace,
                ),
                None,
            ),
        ),
        (
            src_dir / "stringification.cpp",
            lambda: (
                cpp_lib.generate_stringification_implementation(
                    symbol_table=context.symbol_table,
                    library_namespace=library_namespace,
                ),
                None,
            ),
        ),
        (
            include_dir / "types.hpp",
            lambda: cpp_lib.generate_types_header(
                symbol_table=verified_ir_table,
                library_namespace=library_namespace,
            ),
        ),
        (
            src_dir / "types.cpp",
            lambda: cpp_lib.generate_types_implementation(
                symbol_table=context.symbol_table,
                spec_impls=context.spec_impls,
                library_namespace=library_namespace,
            ),
        ),
        (
            include_dir / "verification.hpp",
            lambda: cpp_lib.generate_verification_header(
                symbol_table=verified_ir_table,
                spec_impls=context.spec_impls,
                library_namespace=library_namespace,
            ),
        ),
        (
            src_dir / "verification.cpp",
            lambda: cpp_lib.generate_verification_implementation(
                symbol_table=context.symbol_table,
                spec_impls=context.spec_impls,
                library_namespace=library_namespace,
            ),
        ),
        (
            include_dir / "visitation.hpp",
            lambda: (
                cpp_lib.generate_visitation_header(
                    symbol_table=verified_ir_table, library_namespace=library_namespace
                ),
                None,
            ),
        ),
        (
            src_dir / "visitation.cpp",
            lambda: (
                cpp_lib.generate_visitation_implementation(
                    symbol_table=verified_ir_table, library_namespace=library_namespace
                ),
                None,
            ),
        ),
        (
            include_dir / "wstringification.hpp",
            lambda: (
                cpp_lib.generate_wstringification_header(
                    symbol_table=context.symbol_table,
                    library_namespace=library_namespace,
                ),
                None,
            ),
        ),
        (
            src_dir / "wstringification.cpp",
            lambda: (
                cpp_lib.generate_wstringification_implementation(
                    symbol_table=context.symbol_table,
                    library_namespace=library_namespace,
                ),
                None,
            ),
        ),
        (
            include_dir / "xmlization.hpp",
            lambda: (
                cpp_lib.generate_xmlization_header(
                    symbol_table=verified_ir_table, library_namespace=library_namespace
                ),
                None,
            ),
        ),
        (
            src_dir / "xmlization.cpp",
            lambda: cpp_lib.generate_xmlization_implementation(
                symbol_table=context.symbol_table,
                spec_impls=context.spec_impls,
                library_namespace=library_namespace,
            ),
        ),
        (
            test_dir / "common.hpp",
            lambda: (
                cpp_tests.generate_common_header(library_namespace=library_namespace),
                None,
            ),
        ),
        (
            test_dir / "common.cpp",
            lambda: (
                cpp_tests.generate_common_implementation(
                    library_namespace=library_namespace
                ),
                None,
            ),
        ),
        (
            test_dir / "common_examples.hpp",
            lambda: (
                cpp_tests.generate_common_examples_header(
                    symbol_table=context.symbol_table,
                    library_namespace=library_namespace,
                ),
                None,
            ),
        ),
        (
            test_dir / "common_examples.cpp",
            lambda: (
                cpp_tests.generate_common_examples_implementation(
                    symbol_table=context.symbol_table,
                    library_namespace=library_namespace,
                ),
                None,
            ),
        ),
        (
            test_dir / "common_jsonization.hpp",
            lambda: (
                cpp_tests.generate_common_jsonization_header(
                    library_namespace=library_namespace
                ),
                None,
            ),
        ),
        (
            test_dir / "common_jsonization.cpp",
            lambda: (
                cpp_tests.generate_common_jsonization_implementation(
                    library_namespace=library_namespace
                ),
                None,
            ),
        ),
        (
            test_dir / "common_xmlization.hpp",
            lambda: (
                cpp_tests.generate_common_xmlization_header(
                    library_namespace=library_namespace
                ),
                None,
            ),
        ),
        (
            test_dir / "common_xmlization.cpp",
            lambda: (
                cpp_tests.generate_common_xmlization_implementation(
                    library_namespace=library_namespace
                ),
                None,
            ),
        ),
        (
            test_dir / "test_descent_and_descent_once.cpp",
            lambda: (
                cpp_tests.generate_test_descent_and_descent_once_implementation(
                    symbol_table=context.symbol_table,
                    library_namespace=library_namespace,
                ),
                None,
            ),
        ),
        (
            test_dir / "test_jsonization_dispatch.cpp",
            lambda: (
                cpp_tests.generate_test_jsonization_dispatch_implementation(
                    symbol_table=context.symbol_table,
                    library_namespace=library_namespace,
                ),
                None,
            ),
        ),
        (
            test_dir / "test_jsonization_of_concrete_classes.cpp",
            lambda: (
                cpp_tests.generate_test_jsonization_of_concrete_classes_implementation(
                    symbol_table=context.symbol_table,
                    library_namespace=library_namespace,
                ),
                None,
            ),
        ),
        (
            test_dir / "test_revm.cpp",
            lambda: (
                cpp_tests.generate_test_revm_implementation(
                    library_namespace=library_namespace
                ),
                None,
            ),
        ),
        (
            test_dir / "test_stringification_base64.cpp",
            lambda: (
                cpp_tests.generate_test_stringification_base64_implementation(
                    library_namespace=library_namespace,
                ),
                None,
            ),
        ),
        (
            test_dir / "test_stringification_of_enums.cpp",
            lambda: (
                cpp_tests.generate_test_stringification_of_enums_implementation(
                    symbol_table=context.symbol_table,
                    library_namespace=library_namespace,
                ),
                None,
            ),
        ),
        (
            test_dir / "test_verification.cpp",
            lambda: (
                cpp_tests.generate_test_verification_implementation(
                    symbol_table=context.symbol_table,
                    library_namespace=library_namespace,
                ),
                None,
            ),
        ),
        (
            test_dir / "test_wstringification_of_enums.cpp",
            lambda: (
                cpp_tests.generate_test_wstringification_of_enums_implementation(
                    symbol_table=context.symbol_table,
                    library_namespace=library_namespace,
                ),
                None,
            ),
        ),
        (
            test_dir / "test_xmlization_dispatch.cpp",
            lambda: (
                cpp_tests.generate_test_xmlization_dispatch_implementation(
                    symbol_table=context.symbol_table,
                    library_namespace=library_namespace,
                ),
                None,
            ),
        ),
        (
            test_dir / "test_xmlization_of_concrete_classes.cpp",
            lambda: (
                cpp_tests.generate_test_xmlization_of_concrete_classes_implementation(
                    symbol_table=context.symbol_table,
                    library_namespace=library_namespace,
                ),
                None,
            ),
        ),
        (
            test_dir / "test_x_or_default.cpp",
            lambda: (
                cpp_tests.generate_test_x_or_default_implementation(
                    symbol_table=context.symbol_table,
                    library_namespace=library_namespace,
                ),
                None,
            ),
        ),
    ]

    for rel_path, generator_func in rel_paths_generators:
        assert not rel_path.is_absolute()

        code, errors = generator_func()

        if errors is not None:
            run.write_error_report(
                message=f"Failed to generate {rel_path} "
                f"based on {context.model_path}",
                errors=[
                    context.lineno_columner.error_message(error) for error in errors
                ],
                stderr=stderr,
            )
            return 1

        assert code is not None

        pth = context.output_dir / rel_path

        try:
            pth.parent.mkdir(parents=True, exist_ok=True)
        except Exception as exception:
            run.write_error_report(
                message=f"Failed to create the directory {pth.parent}",
                errors=[str(exception)],
                stderr=stderr,
            )
            return 1

        try:
            pth.write_text(code, encoding="utf-8")
        except Exception as exception:
            run.write_error_report(
                message=f"Failed to write to {pth}",
                errors=[str(exception)],
                stderr=stderr,
            )
            return 1

    stdout.write(f"Code generated to: {context.output_dir}\n")
    return 0


assert cpp.__doc__ == __doc__



"""Generate the constants shared across the unit tests."""

import io
from typing import List

from icontract import ensure

from aas_core_codegen.common import (
    Stripped,
)
from aas_core_codegen.golang import common as golang_common
from aas_core_codegen.golang.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
    INDENT4 as IIII,
)


# fmt: off
@ensure(
    lambda result: result.endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate(repo_url: Stripped) -> str:
    """Generate the constants shared across the unit tests."""
    environment_variable = golang_common.repo_url_to_environment_variable(repo_url)

    blocks = [
        Stripped(
            """\
package aastesting"""
        ),
        golang_common.WARNING,
        Stripped(
            f"""\
import (
{I}"fmt"
{I}"os"
{I}"strings"
)"""
        ),
        Stripped(
            f"""\
const RecordModeEnvironmentVariableName string = (
{I}"{environment_variable}_TEST_RECORD_MODE")"""
        ),
        Stripped(
            '''\
// NOTE (mristin):
// It is tedious to record manually all the expected error messages. Therefore we
// include this variable to steer the automatic recording. We intentionally
// intertwine the recording code with the test code to keep them close to each other
// so that they are easier to maintain.
var rM = os.Getenv(RecordModeEnvironmentVariableName)
var RecordMode = rM == "1" || strings.ToLower(rM) == "true" || strings.ToLower(rM) == "on"'''
        ),
        Stripped(
            f"""\
const TestDataDirEnvironmentVariableName string = (
{I}"{environment_variable}_TEST_DATA_DIR")"""
        ),
        Stripped(
            f"""\
func getTestDataDir() string {{
{I}variable := TestDataDirEnvironmentVariableName
{I}val, ok := os.LookupEnv(variable)
{I}if !ok {{
{II}panic(
{III}fmt.Sprintf(
{IIII}"Expected the environment variable to be set, but it was not: %s",
{IIII}variable,
{III}),
{II})
{I}}}
{I}return val
}}"""
        ),
        Stripped(
            """\
var TestDataDir = getTestDataDir()"""
        ),
        golang_common.WARNING,
    ]  # type: List[Stripped]

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(block)

    writer.write("\n")

    return writer.getvalue()


assert generate.__doc__ is not None
assert generate.__doc__.strip().startswith(__doc__.strip())



"""Generate code to test the XML de/serialization of concrete classes."""

import io
from typing import List

from icontract import ensure

from aas_core_codegen import intermediate, naming
from aas_core_codegen.common import Stripped, Identifier
from aas_core_codegen.typescript import (
    common as typescript_common,
    naming as typescript_naming,
)
from aas_core_codegen.typescript.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
    INDENT4 as IIII,
)


# fmt: off
@ensure(
    lambda result: result.endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate(symbol_table: intermediate.SymbolTable) -> str:
    """Generate code to test the XML de/serialization of concrete classes."""
    blocks = [
        Stripped(
            """\
/**
 * Test XML de/serialization of concrete classes.
 */"""
        ),
        typescript_common.WARNING,
        Stripped(
            """\
import * as fs from "fs";
import * as path from "path";

import * as AasXmlization from "../src/xmlization";
import * as AasTypes from "../src/types";
import * as AasVerification from "../src/verification";

import * as TestCommon from "./common";"""
        ),
    ]  # type: List[Stripped]

    for concrete_cls in symbol_table.concrete_classes:
        cls_name_typescript = typescript_naming.class_name(concrete_cls.name)
        cls_name_xml = naming.xml_class_name(concrete_cls.name)
        as_function = typescript_naming.function_name(
            Identifier(f"as_{concrete_cls.name}")
        )

        blocks.append(
            Stripped(
                f"""\
test("{cls_name_typescript} XML round-trip OK", () => {{
{I}const pths = Array.from(
{II}TestCommon.findFilesBySuffixRecursively(
{III}path.join(
{IIII}TestCommon.TEST_DATA_DIR,
{IIII}"Xml",
{IIII}"Expected",
{IIII}{typescript_common.string_literal(cls_name_xml)}
{III}),
{III}".xml"
{II})
{I});
{I}pths.sort();

{I}for (const pth of pths) {{
{II}const text = fs.readFileSync(pth, "utf-8");

{II}const instanceOrError = AasXmlization.fromXmlString(text);
{II}expect(instanceOrError.error).toBeNull();
{II}const instance = instanceOrError.mustValue();

{II}const casted = AasTypes.{as_function}(instance);
{II}if (casted === null) {{
{III}throw new Error(
{IIII}`Expected instance of {cls_name_typescript} in ${{pth}}, ` +
{IIII}`but got: ${{typeof instance}}`
{III});
{II}}}

{II}TestCommon.assertNoVerificationErrors(AasVerification.verify(casted), pth);

{II}const roundTripText = AasXmlization.toXmlString(casted);
{II}expect(roundTripText.length).toBeGreaterThan(0);
{I}}}
}});"""
            )
        )

        blocks.append(
            Stripped(
                f"""\
test("{cls_name_typescript} XML deserialization fail", () => {{
{I}for (
{II}const causeDir of
{II}TestCommon.findImmediateSubdirectories(
{III}path.join(
{IIII}TestCommon.TEST_DATA_DIR,
{IIII}"Xml",
{IIII}"Unexpected",
{IIII}"Unserializable"
{III})
{II})
{I}) {{
{II}const clsDir = path.join(
{III}causeDir,
{III}{typescript_common.string_literal(cls_name_xml)}
{II});
{II}if (!fs.existsSync(clsDir)) {{
{III}continue;
{II}}}

{II}const pths = Array.from(
{III}TestCommon.findFilesBySuffixRecursively(
{IIII}clsDir,
{IIII}".xml"
{III})
{II});
{II}pths.sort();

{II}for (const pth of pths) {{
{III}const text = fs.readFileSync(pth, "utf-8");
{III}const instanceOrError = AasXmlization.fromXmlString(text);
{III}expect(instanceOrError.error).not.toBeNull();
{II}}}
{I}}}
}});"""
            )
        )

    blocks.append(typescript_common.WARNING)

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(block)

    writer.write("\n")

    return writer.getvalue()


assert generate.__doc__ is not None
assert generate.__doc__.strip().startswith(__doc__.strip())



"""Generate code to test the ``xOrDefault`` methods."""

import io
from typing import List, Optional

from icontract import ensure

from aas_core_codegen import intermediate, naming
from aas_core_codegen.common import Stripped, Identifier, indent_but_first_line
from aas_core_codegen.typescript import (
    common as typescript_common,
    naming as typescript_naming,
)
from aas_core_codegen.typescript.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
    INDENT4 as IIII,
)


# fmt: off
@ensure(
    lambda result: result.endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate(symbol_table: intermediate.SymbolTable) -> str:
    """Generate code to test the ``xOrDefault`` methods."""
    blocks = [
        Stripped(
            """\
/**
 * Test `*OrDefault` functions.
 */"""
        ),
        typescript_common.WARNING,
        Stripped(
            """\
import * as path from "path";
import * as fs from "fs";

import * as AasStringification from "../src/stringification";
import * as AasTypes from "../src/types";
import * as TestCommon from "./common";
import * as TestCommonJsonization from "./commonJsonization";"""
        ),
        Stripped(
            f"""\
/**
 * Represent explicitly a literal of an enumeration.
 */
class EnumerationLiteral {{
{I}constructor(public enumerationName: string, public literalName) {{
{II}// Intentionally empty.
{I}}}

{I}toString(): string {{
{II}return `${{this.enumerationName}}.${{this.literalName}}`;
{I}}}
}}"""
        ),
        Stripped(
            f"""\
/**
 * Represent a value such that we can immediately check whether it is the default value
 * or the set one.
 *
 * @remark
 * We compare it against the recorded golden file, if not {{@link common.RECORD_MODE}}.
 * Otherwise, when {{@link common.RECORD_MODE}} is set, we re-record the golden file.
 *
 * @param value - to be represented
 * @param expectedPath - to the golden file
 */
function compareOrRecordValue(
{I}value: boolean | number | string | null | EnumerationLiteral | AasTypes.Class,
{I}expectedPath: string
): void {{
{I}let got = "";
{I}if (
{II}typeof value === "boolean"
{II}|| typeof value === "number"
{II}|| typeof value === "string"
{II}|| value === null
{I}) {{
{II}got = JSON.stringify(value);
{I}}} else if (value instanceof EnumerationLiteral) {{
{II}got = value.toString();
{I}}} else if (value instanceof AasTypes.Class) {{
{II}got = TestCommon.traceMark(value);
{I}}} else {{
{II}throw new Error(`We do not know how to represent the value ${{value}}`);
{I}}}

{I}// NOTE (mristin):
{I}// We add a new line for POSIX systems which prefer a new line
{I}// at the end of the file.
{I}got += "\\n";

{I}if (TestCommon.RECORD_MODE) {{
{II}const parent = path.dirname(expectedPath);
{II}if (!fs.existsSync(parent)) {{
{III}fs.mkdirSync(parent, {{recursive: true}});
{II}}}
{II}fs.writeFileSync(expectedPath, got, "utf-8");
{I}}} else {{
{II}if (!fs.existsSync(expectedPath)) {{
{III}throw new Error(
{IIII}`The file with the recorded value does not exist: ${{expectedPath}}; ` +
{IIII}`you probably want to set the environment ` +
{IIII}`variable ${{TestCommon.RECORD_MODE_ENVIRONMENT_VARIABLE_NAME}}?`
{III});
{II}}}

{II}const expected =
{III}fs.readFileSync(expectedPath, "utf-8")
{III}.replace(/\\r\\n/g, "\\n");
{II}expect(got).toStrictEqual(expected);
{I}}}
}}"""
        ),
    ]  # type: List[Stripped]

    for concrete_cls in symbol_table.concrete_classes:
        cls_name_typescript = typescript_naming.class_name(concrete_cls.name)
        cls_name_json = naming.json_model_type(concrete_cls.name)

        x_or_default_methods = []  # type: List[intermediate.MethodUnion]
        for method in concrete_cls.methods:
            if method.name.endswith("_or_default"):
                x_or_default_methods.append(method)

        for method in x_or_default_methods:
            method_name_typescript = typescript_naming.method_name(method.name)

            result_enum = None  # type: Optional[intermediate.Enumeration]

            assert method.returns is not None, (
                f"Expected all X_or_default to return something, "
                f"but got None for {concrete_cls}.{method.name}"
            )

            if isinstance(
                method.returns, intermediate.OurTypeAnnotation
            ) and isinstance(method.returns.our_type, intermediate.Enumeration):
                result_enum = method.returns.our_type

            if result_enum is None:
                value_assignment_snippet = Stripped(
                    f"const value = instance.{method_name_typescript}();"
                )
            else:
                enum_to_string_name = typescript_naming.function_name(
                    Identifier(f"must_{result_enum.name}_to_string")
                )

                value_assignment_snippet = Stripped(
                    f"""\
const value = new EnumerationLiteral(
{I}{typescript_common.string_literal(typescript_naming.enum_name(result_enum.name))},
{I}AasStringification.{enum_to_string_name}(
{II}instance.{method_name_typescript}()
{I})
);"""
                )

            load_maximal_name = typescript_naming.function_name(
                Identifier(f"load_maximal_{concrete_cls.name}")
            )

            # noinspection SpellCheckingInspection
            blocks.append(
                Stripped(
                    f"""\
test("{cls_name_typescript}.{method_name_typescript} on maximal", () => {{
{I}const instance = TestCommonJsonization.{load_maximal_name}();

{I}{indent_but_first_line(value_assignment_snippet, I)}

{I}compareOrRecordValue(
{II}value,
{II}path.join(
{III}TestCommon.TEST_DATA_DIR,
{III}"xOrDefault",
{III}{typescript_common.string_literal(cls_name_json)},
{III}"{method_name_typescript}.on_maximal.json"
{II})
{I});
}});"""
                )
            )

            load_minimal_name = typescript_naming.function_name(
                Identifier(f"load_minimal_{concrete_cls.name}")
            )

            blocks.append(
                Stripped(
                    f"""\
test("{cls_name_typescript}.{method_name_typescript} on minimal", () => {{
{I}const instance = TestCommonJsonization.{load_minimal_name}();

{I}{indent_but_first_line(value_assignment_snippet, I)}

{I}compareOrRecordValue(
{II}value,
{II}path.join(
{III}TestCommon.TEST_DATA_DIR,
{III}"xOrDefault",
{III}{typescript_common.string_literal(cls_name_json)},
{III}"{method_name_typescript}.on_minimal.json"
{II})
{I});
}});"""
                )
            )

    blocks.append(typescript_common.WARNING)

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(block)

    writer.write("\n")

    return writer.getvalue()


assert generate.__doc__ is not None
assert generate.__doc__.strip().startswith(__doc__.strip())



"""Generate code to test the verification of enums."""

from typing import List

from aas_core_codegen import intermediate
from aas_core_codegen.common import Stripped, indent_but_first_line
from aas_core_codegen.java import common as java_common, naming as java_naming
from aas_core_codegen.java.common import (
    INDENT as I,
    INDENT2 as II,
)


def generate(
    package: java_common.PackageIdentifier,
    symbol_table: intermediate.SymbolTable,
) -> List[java_common.JavaFile]:
    """
    Generate code to test the verification of enums.
    """
    blocks = []  # type: List[Stripped]

    for enumeration in symbol_table.enumerations:
        enum_name = java_naming.enum_name(enumeration.name)

        assert (
            len(enumeration.literals) > 0
        ), f"Unexpected enumeration without literals: {enumeration.name}"

        literal_name = java_naming.enum_literal_name(enumeration.literals[0].name)

        blocks.append(
            Stripped(
                f"""\
@Test
public void test{enum_name}Valid() {{
{I}final List<Reporting.Error> errors = Verification.verify{enum_name}(
{II}{enum_name}.{literal_name}).collect(Collectors.toList());

{I}assertEquals(0, errors.size());
}} // void test{enum_name}Valid"""
            )
        )

        blocks.append(
            Stripped(
                f"""\
@Test
public void test{enum_name}Invalid() {{
{I}final {enum_name} value = null;
{I}final List<Reporting.Error> errors =
{II}Verification.verify{enum_name}(value).collect(Collectors.toList());

{I}assertEquals(1, errors.size());
{I}assertEquals("Invalid {enum_name}: null", errors.get(0).getCause());
}} // void test{enum_name}Invalid"""
            )
        )

    blocks_joined = "\n\n".join(blocks)

    return [
        java_common.JavaFile(
            "TestVerificationOfEnums.java",
            f"""\
{java_common.WARNING}

package {package}.tests;

import static org.junit.jupiter.api.Assertions.assertEquals;

import {package}.reporting.Reporting;
import {package}.types.enums.*;
import {package}.verification.Verification;
import java.util.List;
import java.util.stream.Collectors;
import org.junit.jupiter.api.Test;

public class TestVerificationOfEnums {{
{I}{indent_but_first_line(blocks_joined, I)}
}} // class TestVerificationOfEnums

// package {package}.tests

{java_common.WARNING}
""",
        )
    ]


assert generate.__doc__ is not None
assert generate.__doc__.strip().startswith(__doc__.strip())



"""Generate code to test the JSON de/serialization of interfaces."""

from typing import List

from aas_core_codegen import intermediate
from aas_core_codegen.common import Stripped, indent_but_first_line
from aas_core_codegen.java import common as java_common, naming as java_naming
from aas_core_codegen.java.common import INDENT as I, INDENT2 as II


def generate(
    package: java_common.PackageIdentifier,
    symbol_table: intermediate.SymbolTable,
) -> List[java_common.JavaFile]:
    """
    Generate code to test the JSON de/serialization of interfaces.
    """
    blocks = []  # type: List[Stripped]

    for cls in symbol_table.classes:
        if cls.interface is None or len(cls.interface.implementers) == 0:
            continue

        for implementer_cls in cls.interface.implementers:
            if (
                implementer_cls.serialization is None
                or not implementer_cls.serialization.with_model_type
            ):
                continue

            interface_name_java = java_naming.interface_name(cls.interface.name)
            implementer_cls_name_java = java_naming.class_name(implementer_cls.name)

            blocks.append(
                Stripped(
                    f"""\
@Test
public void testRoundTrip{interface_name_java}From{implementer_cls_name_java}()
{I}throws IOException, JsonProcessingException {{
{I}final ObjectMapper objectMapper = new ObjectMapper();

{I}{implementer_cls_name_java} instance = CommonJsonization.loadMaximal{implementer_cls_name_java}();
{I}final JsonNode jsonObject = Jsonization.Serialize.toJsonObject(instance);

{I}{interface_name_java} anotherInstance = Jsonization.Deserialize.deserialize{interface_name_java}(
{II}jsonObject);
{I}final JsonNode anotherJsonObject = Jsonization.Serialize.toJsonObject(anotherInstance);

{I}assertEquals(
{II}objectMapper.readTree(jsonObject.toString()), objectMapper.readTree(anotherJsonObject.toString()));
}} // void testRoundTrip{interface_name_java}From{implementer_cls_name_java}"""
                )
            )

    blocks_joined = "\n\n".join(blocks)

    return [
        java_common.JavaFile(
            "TestJsonizationOfInterfaces.java",
            f"""\
{java_common.WARNING}

package {package}.tests;

import static org.junit.jupiter.api.Assertions.assertEquals;

import {package}.jsonization.Jsonization;
import {package}.reporting.*;
import {package}.types.impl.*;
import {package}.types.model.*;
import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.JsonNode;
import com.fasterxml.jackson.databind.ObjectMapper;
import java.io.IOException;
import org.junit.jupiter.api.Test;

public class TestJsonizationOfInterfaces {{
{I}{indent_but_first_line(blocks_joined, I)}
}} // class TestJsonizationOfInterfaces

// package {package}.tests

{java_common.WARNING}
""",
        )
    ]


assert generate.__doc__ is not None
assert generate.__doc__.strip().startswith(__doc__.strip())
